"""
Package data
"""
import os

def prompts():
    """
    Scrapped from RCSB PDB website- use for creating query prompts
    :return: dict
    """
    return {"HoldingsQuery": "All/Experimental Type/Molecule Type", "StructureIdQuery": "PDB ID(s)",
            "EntityIdQuery": "Entity ID(s)", "ChainIdQuery": "Chain ID(s)", "PubmedIdQuery": "PubMed ID(s)",
            "UpAccessionIdQuery": "UniProtKB Accession Number(s)", "AdvancedKeywordQuery": "Text Search",
            "TokenKeywordQuery": "mmCIF Keyword Search (Classification)", "PfamIdQuery": "Pfam Accession Number(s)",
            "UniprotGeneNameQuery": "UniProt Gene Name", "SequenceClusterQuery": "Sequence Cluster Name",
            "StructTitleQuery": "Structure Title", "StructDescQuery": "Structure Description",
            "MoleculeNameQuery": "Macromolecule Name", "PathwayQuery": "Pathway Reaction Name",
            "LargeStructureQuery": "Large Structures", "AdvancedAuthorQuery": "Author Name",
            "DepositDateQuery": "Deposit Date", "ReleaseDateQuery": "Release Date", "ReviseDateQuery": "Revision Date",
            "LastLoadQuery": "Latest Released Structures", "ModifiedStructuresQuery": "Latest Modified Structures",
            "SGProjectQuery": "Structural Genomics Project", "ChainTypeQuery": "Macromolecule Type",
            "NumberOfChainsQuery": "Number of Chains (Asymmetric Unit)",
            "BiolUnitQuery": "Number of Chains (Biological Assembly)", "NumberOfEntitiesQuery": "Number of Entities",
            "StoichiometryQuery": "Protein Stoichiometry", "PointGroupQuery": "Protein Symmetry",
            "TreeQuery_20": "Protein Symmetry Browser", "ModelCountQuery": "Number of Models",
            "CloseContactsQuery": "Number of Disulfide Bonds", "LinkConnectionQuery": "Link records",
            "MolecularWeightQuery": "Molecular Weight (Structure)",
            "SecondaryStructureQuery": "Secondary Structure Content",
            "SecondaryStructureLengthQuery": "Secondary Structure Length",
            "TreeQuery_11": "SCOP Classification Browser",
            "TreeQuery_12": "CATH Classification Browser", "TreeQueryExpression_1": "Taxonomy Browser",
            "SequenceQuery": "Sequence (BLAST/PSI-BLAST)", "WildTypeProteinQuery": "Wild Type Protein",
            "MutationQuery": "Mutation", "BlastXQuery": "Translated Nucleotide Sequence (BLASTX)",
            "MotifQuery": "Sequence Motif", "SequenceLengthQuery": "Chain Length",
            "ProteinModificationsQuery": "Protein Modifications", "TreeQuery_17": "Protein Modification Browser",
            "TreeEntityQuery_13": "Genome Location Browser", "ChemCompNameQuery": "Chemical Name",
            "ChemCompIdQuery": "Chemical ID(s)", "ChemCompDescriptorQuery": "InChI Descriptor",
            "ChemSmilesQuery": "Chemical structure (SMILES)",
            "ChemFormulaWeightQuery": "Molecular Weight (Chemical component)",
            "ChemCompFormulaQuery": "Chemical Formula", "ChemCompTypeQuery": "Chemical Component Type",
            "BindingAffinityQuery": "Binding Affinity", "NoLigandQuery": "Has Ligand(s)",
            "NoModResQuery": "Has Modified Residue(s)", "ChemCompSubCompQuery": "Sub-components",
            "BirdQuery": "Biologically Interesting Molecules (from BIRD)",
            "TreeEntityQuery_1": "Source Organism Browser (NCBI)", "ExpressionOrganismQuery": "Expression Organism",
            "TreeEntityQuery_3": "Enzyme Classification Browser", "EnzymeClassificationQuery": "Enzyme Classification",
            "TreeEntityQuery_4": "Biological Process Browser (GO)", "TreeEntityQuery_5": "Cell Component Browser (GO)",
            "TreeEntityQuery_6": "Molecular Function Browser (GO)",
            "TreeQuery_16": "Transporter Classification Browser",
            "ExpTypeQuery": "Experimental Method", "ResolutionQuery": "X-ray Resolution",
            "AverageBFactorQuery": "X-ray Average B Factor", "XrayRefinementQuery": "Refinement R Factors",
            "XrayDiffrnSourceQuery": "Diffraction Source",
            "MethodToDetermineStructQuery": "Structure Determination Method", "XrayReflnsQuery": "Reflections",
            "XrayCellQuery": "Cell Dimensions", "SoftwareQuery": "Software", "SpaceGroupQuery": "Space Group",
            "CrystalQuery": "Crystal Properties", "EmAssemblyQuery": "EM Assembly", "DetectorTypeQuery": "Detector",
            "CitationTwoQuery": "Citation", "TreeQuery_15": "Medical Subject Headings Browser",
            "PubMedQuery": "PubMed Abstract", "ExternalLinkQuery": "Has External Links"}


def common_solvents():
    """
    common solvents/ other junk in PDB entries
    :return:
    """
    f = open(os.path.join(os.path.split(__file__)[0], "excluded_het_id.txt"), 'r')
    return f.read().splitlines()



from hotspots import calculation


__author__ = "Chris Radoux, Peter Curran, Mihaela Smilova"
__copyright__ = None
__credits__ = None
__license__ = None
__version__ = "1.0.0"
__maintainer__ = "Peter Curran"
__email__ = "pcurran@ccdc.cam.ac.uk"
__status__ = "Development"
name = "hotspots"



"""
python wrapper for the rcsb PDB

Thank-you pypdb.py for a great starting point [ref]

- aiming to open up all search combinations
- composite queries
"""
from __future__ import print_function, division

import os
import sys
import shutil

import xmltodict
from data import common_solvents
from tqdm import tqdm

if sys.version_info.major == 2:
    from urllib2 import Request as request
    from urllib2 import urlopen as urlopen

else:
    from urllib.request import Request as request
    from urllib.request import urlopen as urlopen


base_url = 'http://www.rcsb.org/pdb/rest'


class Helper(object):
    """
    generic functionality to aid RESTful python splender
    """

    @staticmethod
    def url_request(extension, data=None):
        """"""
        if data:
            url = request("{}/{}".format(base_url, extension), data=data)
        else:
            url = request("{}/{}".format(base_url, extension))

        return urlopen(url)


class Query(Helper):
    """
    constructs a query
    """

    class _QueryTerm(object):
        """

        """
        def __init__(self, query_type, query_parameters, conjunction):
            self.query_parameters = query_parameters
            self.conjunction = conjunction
            self.prefix = "org.pdb.query.simple."

            self.query_type = self.prefix + query_type
            self.query_parameters["queryType"] = self.query_type

            self._generate_description()
            self.term = {'orgPdbQuery': self.query_parameters}

        def _generate_description(self):
            """

            :return:
            """
            self.query_parameters["description"] = 'No description'

    def __init__(self):
        self.query_terms = []

    @property
    def query(self):
        """
        Assembles the QueryTerms added into correct dictionary format
        :return:
        """
        query_term_dict = {"queryRefinement": []}
        for i, query_term in enumerate(self.query_terms):

            if i > 0:
                try:
                    query_term.term.update({"conjunctionType": query_term.conjunction})
                except:
                    raise ValueError("For Queries with > 1 QueryTerm, conjuction can not be NoneType")

            query_term.term.update({"queryRefinementLevel": str(i)})
            query_term_dict["queryRefinement"].append(query_term.term)

        return {"orgPdbCompositeQuery": query_term_dict}

    def add_term(self, query_type, query_parameters, conjunction=None):
        """

        :param query_type:
        :param query_parameters:
        :return:
        """
        self.query_terms.append(Query._QueryTerm(query_type, query_parameters, conjunction))


class PDB(object):
    """
    this is a toolkit to access the PDB's RESTful web service.
    """
    @staticmethod
    def search(query=None):
        """
        search the pdb, if query is None the whole PDB will be returned
        :param query:
        :return:
        """
        print("Searching PDB ...")
        if query:
            query_text = xmltodict.unparse(query, pretty=False)
            query_text = query_text.encode()
            f = Helper.url_request(extension='/search', data=query_text)

            return [PDBResult(identifier=code.split(":")[0]) for code in tqdm(f.readlines())]

        else:
            f = Helper.url_request(extension='/getCurrent')
            f.read()


class _Ligand(Helper):
    """
    a class to handle PDB ligands (keep at string level)
    """

    def __init__(self, ligand_info):
        """

        :param ligand_str:
        """
        self.ligand_info = ligand_info

    @staticmethod
    def from_chemicalid(chemicalid):
        """

        :param chemicalid:
        :return:
        """

        extension = "describeHet?chemicalID={}".format(chemicalid)
        data = xmltodict.parse(Helper.url_request(extension=extension),
                               attr_prefix='',
                               dict_constructor=dict)

        return _Ligand(ligand_info=data.values()[0].values()[0]['ligand'])


    def search(self, search_type, tanimoto=0.7):
        """
        will do a chemical structure search using _Ligands smile string.
        Search Types:
            - 'exact'
            - 'substructure'
            - 'superstructure'
            - 'similarity'

        :param search_type: type of search to be executed
        :param tanimoto: float, tanimoto cooefficent for similarity search. 0 = most dissimilar, 1 = identifical
        :return:
        """
        supported = ["exact", "substructure", "superstructure", "similarity"]
        assert search_type in supported

        extension = "smilesQuery?smiles={0}&search_type={1}".format(self.smiles, search_type)
        if search_type is 'similarity':
            extension += "&similarity={}".format(tanimoto)

        data = xmltodict.parse(self.url_request(extension=extension),
                               attr_prefix='',
                               dict_constructor=dict)

        for key, value in data.values()[0].items():
            if key == "ligandInfo":
                return [_Ligand(ligand_info=v) for v in value.values()[0]]

    @property
    def structure_id(self):
        return self.ligand_info["structureId"]

    @property
    def smiles(self):
        return self.ligand_info["smiles"]

    @property
    def InChIKey(self):
        return self.ligand_info["InChIKey"]

    @property
    def molecular_weight(self):
        return self.ligand_info["molecularWeight"]

    @property
    def chemical_id(self):
        return self.ligand_info["chemicalID"]

    @property
    def chemical_name(self):
        return self.ligand_info["chemicalName"]

    @property
    def formula(self):
        return self.ligand_info["formula"]

    @property
    def chemical_type(self):
        return self.ligand_info["type"]


class _Protein(object):
    """
    """
    class _SubUnit(object):
        def __init__(self, data):
            self.data = data

        @property
        def entity_number(self):
            return self.data["entityNr"]

        @property
        def chain(self):
            return self.data["chain"]["id"]

        @property
        def accession_id(self):
            return self.data["macroMolecule"]["accession"]["id"]

    def __init__(self, data):
        """
        :param ligand_str:
        """
        self.data = data
        self.sub_unit_data = []
        for data_item in self.data.values():
            for i in data_item.values():
                if type(i) is dict:
                    self.sub_unit_data.append(i)
                elif type(i) is unicode:
                    self.structure_id = i
                elif type(i) is list:
                    self.sub_unit_data.append(i[0])
                else:
                    raise TypeError("{}".format(i))

    def sub_unit_count(self):
        return len(self.sub_unit_data)

    @property
    def sub_units(self):
        return [self._SubUnit(sud) for sud in (self.sub_unit_data)]


class _EntryDescription(object):
    """"""
    def __init__(self, data):
        """

        :param ligand_str:
        """
        self.data = data

    @property
    def status(self):
        return self.data["status"]

    @property
    def residue_count(self):
        return self.data["nr_residues"]

    @property
    def structure_authors(self):
        return self.data["structure_authors"]

    @property
    def citation_authors(self):
        return self.data["citation_authors"]

    @property
    def resolution(self):
        return self.data["resolution"]

    @property
    def title(self):
        return self.data["title"]

    @property
    def release_date(self):
        return self.data["release_date"]

    @property
    def experiment_method(self):
        return self.data["expMethod"]

    @property
    def pubmedId(self):
        return self.data["pubmedId"]

    @property
    def keywords(self):
        return self.data["keywords"]

    @property
    def atom_count(self):
        return self.data["nr_atoms"]

    @property
    def entities(self):
        return self.data["nr_entities"]

    @property
    def last_modification_date(self):
        return self.data["last_modification_date"]

    @property
    def deposition_date(self):
        return self.data["deposition_date"]

    @property
    def related_structures(self):
        return [d["pdbId"] for d in self.data["relatedPDB"]]


class PDBResult(object):
    """
    class to handle operations on a single PDB code
    """

    def __init__(self, identifier):
        """

        :param identifier:
        """
        self.identifier = identifier

        self._ligands = self.get_ligands()

    def _raw_properties(self, info_type='ligand'):
        """

        :return:
        """
        info_type_dict = {'describe_pdb': '/describePDB?structureId=',
                          'describe_mol': '/describeMol?structureId=',
                          'ligand': '/ligandInfo?structureId=',
                          'pfam': '/hmmer?structureId=',
                          'general': '/getEntityInfo?structureId='}

        url = request(base_url + info_type_dict[info_type] + self.identifier)
        return urlopen(url)

    @property
    def descriptions(self):
        """returns description"""
        data = xmltodict.parse(self._raw_properties(info_type='describe_pdb').read(),
                               attr_prefix='',
                               dict_constructor=dict)
        for pdb_description in data.values():
            data_item = pdb_description.values()[0]

        return _EntryDescription(data_item)

    @property
    def protein(self):
        """"""
        data = xmltodict.parse(self._raw_properties(info_type='describe_mol').read(),
                               attr_prefix='',
                               dict_constructor=dict)
        return _Protein(data.values()[0])

    @property
    def ligands(self):
        """"""
        return self._ligands

    def get_ligands(self):
        """returns Ligand"""
        data = xmltodict.parse(self._raw_properties(info_type='ligand').read(),
                               attr_prefix='',
                               dict_constructor=dict)

        ligs = []
        for structureId in data.values():
            for ligandInfo in structureId.items():
                for ligand in ligandInfo:
                    if type(ligand) is dict:
                        for data_item in ligand.values():
                            if type(data_item) == list:
                                for d in data_item:
                                    ligs.append(_Ligand(d))
                            else:
                                ligs.append(_Ligand(data_item))
        return ligs

    @property
    def filtered_ligands(self):
        """"""
        cs = common_solvents()
        return [lig for lig in self._ligands if lig.chemical_id not in cs and
                                                float(lig.molecular_weight) > 120]

    def download(self, out_dir, compressed=False, biological_assembly=False):
        """
        only PDB for now, I will put the rest in later #agile
        :param compressed:
        :param biological_assembly
        :return:
        """
        if biological_assembly:
            extension = ".pdb1"

        else:
            extension = ".pdb"

        url = "https://files.rcsb.org/download/{}{}".format(self.identifier, extension)
        pdbfile = urlopen(url).read()

        if compressed:
            out = os.path.join(out_dir, self.identifier)
            if not os.path.exists(out):
                os.mkdir(out)

            self.fname = os.path.join(out, self.identifier + extension)

            with open(self.fname, "w") as writer:
                writer.write(pdbfile)

            shutil.make_archive(out, 'zip', out)
            shutil.rmtree(out)

        else:
            self.fname = os.path.join(out_dir, self.identifier + extension)
            with open(self.fname, "w") as writer:
                writer.write(pdbfile)





"""
The :mod:`hotspots.hs_docking` module contains functionality which
faciliates the **automatic** application of insights from Fragment
Hotspot Maps to docking.


This module is designed to extend the existing CSD python API

More information about the CSD python API is available:
    - The Cambridge Structural Database C.R. Groom, I. J. Bruno, M. P. Lightfoot and S. C. Ward, Acta Crystallographica Section B, B72, 171-179, 2016 [DOI: 10.1107/S2052520616003954]
    - CSD python API 2.0.0 `documentation <https://downloads.ccdc.cam.ac.uk/documentation/API/>`_

More information about the GOLD method is available:
    - Development and Validation of a Genetic Algorithm for Flexible Docking G. Jones, P. Willett, R. C. Glen, A. R. Leach and R. Taylor, J. Mol. Biol., 267, 727-748, 1997 [DOI: 10.1006/jmbi.1996.0897]
"""
from __future__ import print_function
import os
import tempfile

import numpy as np
from ccdc import docking
from ccdc import io
from ccdc import molecule
from ccdc.utilities import _private_importer
from hotspots.result import Extractor, Results

with _private_importer():
    import DockingLib

DockingLib.licence_check()


class DockerSettings(docking.Docker.Settings):
    """
    A class to handle the integration of Fragment Hotspot Map data with GOLD

    This class is designed to mirror the existing CSD python API for smooth integration. For use, import this class
    as the docking settings rather than directly from the Docking API.

    >>> from ccdc.docking import Docker
    >>> from ccdc.protein import Protein
    >>> from hotspots.calculation import Runner
    >>> from hotspots.hs_docking import DockerSettings

    >>> protein = Protein.from_file("1hcl.pdb")

    >>> runner = Runner()
    >>> hs = runner.from_protein(protein)

    >>> docker.settings.add_protein_file("1hcl.pdb")
    >>> docker.settings.add_ligand_file("dock_me.mol2", ndocks=25)
    >>> constraints = docker.settings.HotspotHBondConstraint.from_hotspot(protein=docker.settings.proteins[0], hr=hs)
    >>> docker.settings.add_constraint(constraints)
    >>> docker.dock()

    >>> docker.Results(docker.settings).ligands

    """
    class HotspotHBondConstraint(docking.Docker.Settings.Constraint):
        """
        A protein HBond constraint constructed from a hotspot
        Assign Protein Hbond constraints based on the highest scoring interactions.

        :param list atoms: list of :class:`ccdc.molecule.Atom` instances from the protein. *NB: The atoms should be donatable hydrogens or acceptor atoms.*
        :param weight: the penalty to be applied for no atom of the list forming an HBond.
        :param min_hbond_score: the minimal score of an HBond to be considered a valid HBond.
        """
        def __init__(self, atoms, weight=5.0, min_hbond_score=0.001, _constraint=None):
            self.atoms = atoms[:]
            for a in self.atoms:
                if not self._is_protein_atom(a):
                    raise RuntimeError('One of the atoms is not in the protein')
                if (
                        not (a.atomic_number == 1 and a.neighbours[0].is_donor) and
                        not (a.is_acceptor)
                ):
                    raise RuntimeError('One of the atoms does not form an HBond')
            self._add_to_protein = atoms[0]._molecule
            self.weight = weight
            self.min_hbond_score = min_hbond_score
            if _constraint is None:
                _constraint = DockingLib.GoldProteinHBondConstraint()
                _constraint.from_string(self._to_string())
            super(self.__class__, self).__init__(_constraint)

        @staticmethod
        def from_file(path, protein, weight, min_hbond_score=0.2, max=2):
            """
            create a hotspot constraint from file

            :return:
            """
            constraints = Results.ConstraintData.read(path)

            return [DockerSettings.HotspotHBondConstraint(atoms=[protein.atoms[index]],
                                                          weight=float(weight),
                                                          min_hbond_score=float(min_hbond_score))
                    for i, index in enumerate(constraints.score_by_index.values()) if i < max]

        @staticmethod
        def create(protein, hr, max_constraints=2, weight=5.0, min_hbond_score=0.001, cutoff=14):
            """
            creates a :class:`hotspots.hs_docking.HotspotHBondConstraint`

            :param `ccdc.protein.Protein` protein: the protein to be used for docking
            :param `hotspots.calculation.Result` hr: a result from Fragment Hotspot Maps
            :param int max_constraints: max number of constraints
            :param float weight: the constraint weight (default to be determined)
            :param float min_hbond_score: float between 0.0 (bad) and 1.0 (good) determining the minimum hydrogen bond quality in the solutions.
            :param cutoff: minimum score required to assign the constraint
            :return list: list of :class:`hotspots.hs_docking.HotspotHBondConstraint`
            """
            for atm in protein.atoms:
                atm.partial_charge = int(0)
            prot = hr.score(protein)

            coords = np.array([a.coordinates for a in prot.atoms])
            atm_dic = {atm.partial_charge: atm for atm in prot.atoms
                       if type(atm.partial_charge) is float
                       and ((atm.atomic_number == 1 and atm.neighbours[0].is_donor) or atm.is_acceptor)
                       and _is_solvent_accessible(coords, atm, min_distance=2)
                       }

            print(atm_dic)

            if len(atm_dic) > max_constraints:
                scores = sorted([f[0] for f in atm_dic.items()], reverse=True)[:max_constraints]
            else:
                scores = sorted([f[0] for f in atm_dic.items()], reverse=True)

            return [DockerSettings.HotspotHBondConstraint(atoms=[atm_dic[s]],
                                                          weight=weight,
                                                          min_hbond_score=min_hbond_score)
                    for s in scores if s > cutoff]

        def _to_string(self):
            s = '%.4f %.4f %s' % (
                self.weight, self.min_hbond_score, ' '.join(str(a.index + 1) for a in self.atoms)
            )
            return s

    def add_fitting_points_from_file(self, path='fit_pts.mol2'):
        """
        creates fitting pts from a .mol2 file

        :param str path: path to fitting pts file
        """
        self.fitting_points_file = path

    def add_fitting_points(self, hr, volume=400, threshold=17, mode='threshold'):
        """
        uses the Fragment Hotspot Maps to generate GOLD fitting points.

        GOLD fitting points are used to help place the molecules into the protein cavity. Pre-generating these fitting
        points using the Fragment Hotspot Maps helps to biast results towards making Hotspot interactions.

        :param `hotspots.result.Result` hr: a Fragment Hotspot Maps result
        :param int volume: volume of the occupied by fitting points in Angstroms ^ 3
        :param float threshold: points above this value will be included in the fitting points
        :param str mode: 'threshold'- assigns fitting points based on a score cutoff or 'bcv'- assigns fitting points from best continuous volume analysis (recommended)
        """
        temp = tempfile.mkdtemp()
        mol = molecule.Molecule(identifier="fitting_pts")

        if mode == 'threshold':
            dic = hr.super_grids["apolar"].grid_value_by_coordinates(threshold=threshold)

        elif mode == 'bcv':
            extractor = Extractor(hr=hr, volume=volume, mode="global")
            bv = extractor.extracted_hotspots[0].best_island
            dic = bv.grid_value_by_coordinates(threshold=17)

        else:
            raise TypeError("{} not supported, see documentation for details".format(mode))

        for score, v in dic.items():
            for pts in v:
                atm = molecule.Atom(atomic_symbol='C',
                                    atomic_number=14,
                                    label='{:.2f}'.format(score),
                                    coordinates=pts)
                atm.partial_charge = score
                mol.add_atom(atom=atm)

        fname = os.path.join(temp, 'fit_pts.mol2')
        with io.MoleculeWriter(fname) as w:
            w.write(mol)

        self.fitting_points_file = fname


def _is_solvent_accessible(protein_coords, atm, min_distance=2):
    """
    given a protein and an atom of a protein, determine if the atom is accessible to solvent

    :param protein:
    :param atm:
    :return:
    """
    if str(atm.atomic_symbol) == 'H':
        atm_position = np.array(atm.coordinates)
        neighbour = np.array(atm.neighbours[0].coordinates)
        direction = np.subtract(atm_position, neighbour) * 2
        position = np.array([direction + atm_position])
        distance = min(np.linalg.norm(protein_coords - position, axis=1))
        if distance > min_distance:
            return True
        else:
            return False

    else:
        return True


docking.Docker.Settings = DockerSettings



"""
The :mod:`hotspots.utilities` module contains classes to for
general functionality.

The main classes of the :mod:`hotspots.extraction` module are:
    - :class:`hotspots.hs_utilities.Helper`
    - :class:`hotspots.hs_utilities.Figures`

"""
from __future__ import division

import collections
import math
import tempfile
from os import mkdir
from os.path import exists, join, abspath

import matplotlib as mpl
mpl.use('TkAgg')
import matplotlib.pyplot as plt

import numpy as np

from ccdc.cavity import Cavity
from ccdc.io import MoleculeWriter
from ccdc.molecule import Molecule, Atom

Coordinates = collections.namedtuple('Coordinates', ['x', 'y', 'z'])


class Helper(object):
    """
    A class to handle miscellaneous functionality
    """
    @staticmethod
    def get_distance(coords1, coords2):
        """
        given two coordinates, calculates the distance

        :param tup coords1: float(x), float(y), float(z), coordinates of point 1
        :param tup coords2: float(x), float(y), float(z), coordinates of point 2
        :return: float, distance
        """
        xd = coords1[0] - coords2[0]
        yd = coords1[1] - coords2[1]
        zd = coords1[2] - coords2[2]
        d = math.sqrt(xd ** 2 + yd ** 2 + zd ** 2)
        return d

    @staticmethod
    def get_out_dir(path):
        """
        checks if directory exists, if not, it create the directory

        :param str path: path to directory
        :return str: path to output directory
        """
        if not exists(abspath(path)):
            mkdir(abspath(path))
        return abspath(path)

    @staticmethod
    def get_lines_from_file(fname):
        """
        gets lines from text file, used in Ghecom calculation

        :return: list, list of str
        """
        f = open(fname)
        lines = f.readlines()
        f.close()
        for i in range(0, len(lines)):
            lines[i] = lines[i].strip()
        return lines

    @staticmethod
    def cavity_centroid(obj):
        """
        returns the centre of a cavity

        :param obj: can be a `ccdc.cavity.Cavity` or
        :return: Coordinate
        """
        if isinstance(obj, Cavity):
            features = [f.coordinates for f in obj.features]

        else:
            features = obj.surface_points

        x_avg = round(np.mean([feat[0] for feat in features if isinstance(feat[0], float)]))
        y_avg = round(np.mean([feat[1] for feat in features if isinstance(feat[1], float)]))
        z_avg = round(np.mean([feat[2] for feat in features if isinstance(feat[2], float)]))

        return Coordinates(x=x_avg, y=y_avg, z=z_avg)

    @staticmethod
    def cavity_from_protein(prot):
        """
        currently the Protein API doesn't support the generation of cavities directly from the Protein instance
        this method handles the tedious writing / reading

        :param `ccdc.protein.Protein` prot: protein
        :return: `ccdc.cavity.Cavity`
        """
        tfile = join(tempfile.mkdtemp(), "protein.pdb")
        with MoleculeWriter(tfile) as writer:
            writer.write(prot)

        return Cavity.from_pdb_file(tfile)

    @staticmethod
    def get_label(input, threshold=None):
        """
        creates a value labels from an input grid dictionary

        :param dic input: key = "probe identifier" and value = `ccdc.utilities.Grid`
        :return `ccdc.molecule.Molecule`: pseduomolecule which contains score labels
        """
        min_size_dict = {"apolar": 40,
                         "donor": 15,
                         "acceptor": 15,
                         "negative": 15,
                         "positive": 15}

        atom_dic = {"apolar": 'C',
                    "donor": 'N',
                    "acceptor": 'O',
                    "negative": 'S',
                    "positive": 'H'}

        try:
            interaction_types = [atom_dic[feat.feature_type] for feat in input._features]
            coordinates = [feat.feature_coordinates for feat in input._features]
            scores = [feat.score_value for feat in input._features]

        except AttributeError:

            print(threshold)
            try:
                if threshold is None:
                    pass
                else:
                    interaction_types = []
                    coordinates = []
                    scores = []
                    for p, g in input.items():
                        for island in g.islands(threshold=threshold):
                            if island.count_grid() > min_size_dict[p]:
                                interaction_types.append(atom_dic[p])
                                coordinates.append(island.centroid())
                                scores.append(island.grid_score(threshold=threshold, percentile=50))
            except AttributeError:
                print("object not supported")

        mol = Molecule(identifier="pharmacophore_model")

        pseudo_atms = [Atom(atomic_symbol=interaction_types[i],
                            atomic_number=14,
                            coordinates=coordinates[i],
                            label=str(scores[i]))
                       for i in range(len(interaction_types))]

        for a in pseudo_atms:
            mol.add_atom(a)
        return mol

    def get_atom_type(self, atom):
        """
        return the atom classification

        :param atom:
        :return:
        """
        if atom.is_donor and atom.is_acceptor:
            return "doneptor"
        elif atom.is_acceptor:
            return "acceptor"
        elif atom.is_donor:
            return "donor"
        else:
            return "apolar"


class Figures(object):
    """
    Class to handle the generation of hotspot related figures

    TO DO: is there a better place for this to live?
    """
    @staticmethod
    def histogram(hr):
        """
        creates a histogram from the hotspot scores

        :param `hotspots.result.Results` hr: a Fragment Hotspot Map result
        :return: data, plot
        """
        data = {}
        for p, g in hr.super_grids.items():
            array = g.get_array()
            masked_array = np.ma.masked_less_equal(array, 1)
            grid_values = masked_array.compressed()
            data.update({p: grid_values})

        plt = Figures._plot_histogram(data)
        return data, plt


    # @staticmethod
    # def _2D_diagram(hr, ligand, fpath, title):
    #     '''
    #     Display the distribution of scores as a heatmap on a 2D depiction of the molecule
    #
    #     :param ligand: a :class:`ccdc.Molecule` object.
    #     :param title: str, Title placed at the top of the image
    #     :param output: str, Output file name
    #     :return:
    #     '''
    # try:
    #     from rdkit import Chem
    #     from rdkit.Chem import Draw
    #     from rdkit.Chem import AllChem
    #     from matplotlib.colors import LinearSegmentedColormap
    # except ImportError:
    #     print("""rdkit is needed for this method""")
    #     exit()
    #
    #     mol = MoleculeReader(ligand)[0]
    #
    #     if ligand.split(".")[-1] == "mol2":
    #         with open(ligand, 'r') as lig:
    #             data = lig.read()
    #         mol_rdkit = Chem.MolFromMol2Block(data)
    #         AllChem.Compute2DCoords(mol_rdkit)
    #
    #     elif ligand.split(".")[-1] == "sdf":
    #         suppl = Chem.SDMolSupplier(ligand)
    #         mol_rdkit = suppl[0]
    #         AllChem.Compute2DCoords(mol_rdkit)
    #     else:
    #         print("Method supports .mol2 files only!")
    #         raise ValueError
    #
    #     scores = hr.score_ligand_atoms(mol, schematic=True, tolerance=2)
    #     num_atoms = mol_rdkit.GetNumAtoms()
    #     a = 0.9 / (float(num_atoms))
    #     s = 0.005
    #
    #     contribs = [float(scores[mol_rdkit.GetAtomWithIdx(i).GetProp('_TriposAtomName')]) for i in
    #                 range(mol_rdkit.GetNumAtoms())]
    #
    #     fig = Draw.MolToMPL(mol_rdkit)
    #
    #     cm = colourmap(scheme="inferno")
    #     test_cm = LinearSegmentedColormap.from_list(__file__, cm)
    #     try:
    #         x, y, z = Draw.calcAtomGaussians(mol_rdkit, a, step=s, weights=contribs)
    #
    #         fig.axes[0].imshow(z, cmap=test_cm, interpolation='bilinear', origin='lower', alpha=0.9,
    #                            extent=(0, 1, 0, 1))
    #         fig.axes[0].contour(x, y, z, 5, colors='k', alpha=0.2)
    #     except ValueError:
    #         print("")
    #
    #     if title:
    #         fig.text(1.25, 2.3, title, fontsize=20, horizontalalignment='center', verticalalignment='top',
    #                  color="white")
    #
    #     fig.savefig(output, bbox_inches='tight')

    @staticmethod
    def _plot_histogram(data, title="Fragment Hotspot Maps"):
            """
            initialise the matplotlib figure to output histograms

            :param data:
            :return:
            """
            colour_dict = {"acceptor": "r",
                           "donor": "b",
                           "apolar": "y",
                           "negative": "m",
                           "positive": "c"}
            plt.figure(1)
            for n, key in enumerate(data.keys()):
                j = int(len(data.keys()))
                plt.subplot(j, 1, (n + 1))
                hist, bin_edges = np.histogram(data[key], bins=range(0, 40), normed=True)
                Figures._histogram_settings(bin_edges)
                if n == 0:
                    plt.title(title)
                if n < j - 1:
                    plt.xticks([])
                if n == j - 1:
                    plt.xlabel("Fragment hotspot score")
                if n == int(round(j / 2) - 1):
                    plt.ylabel("Frequency")
                plt.bar(bin_edges[:-1], hist, width=1, color=colour_dict[key])
            return plt

    @staticmethod
    def _histogram_settings(bin_edges):
        """
        display settings
        :param bin_edges:
        :param title:
        :return:
        """
        plt.xlim(min(bin_edges), max(bin_edges))
        plt.ylim(0, 0.35)
        plt.yticks([])

    def _histogram_info(self, data, key, n):
        """
        data for the matplotlib figure

        :param data:
        :param key:
        :param n:
        :return:
        """

        colour_dict = {"acceptor": "r", "donor": "b", "apolar": "y"}
        hist, bin_edges = np.histogram(data[key], bins=range(0, 40), normed=True)
        plt.bar(bin_edges[:-1], hist, width=1, color=colour_dict[key])
        plt.xlim(min(bin_edges), max(bin_edges))
        plt.ylim(0, 0.35)
        plt.yticks([])
        if n == 0:
            plt.title("Fragment-hotspot Maps")
        if n < 2:
            plt.xticks([])
        if n == 2:
            plt.xlabel("Fragment hotspot score")
        if n == 1:
            plt.ylabel("Frequency")




'''
The main class of the :mod:`hotspots.grid_extension.Grid`.

This is an internal extension of :class:`ccdc.grid.Grid` that adds in potential new _features
for review and development. For example, gaussian smoothing function:

.. code-block:: python
#
# >>> from ccdc_development import Grid
# >>> from scipy import ndimage
# >>> import numpy as np
# >>>
# >>> grd = Grid.from_file(<path/to/file>)
# >>> new_grd = grd.gaussian(sigma=0.3)
# >>>

I.e. this returns a new grid object in which gaussian smoothing has been applied
'''
#############################################################################
from __future__ import print_function, division

import collections
import operator
import numpy as np
from ccdc import utilities
from hotspots.hs_utilities import Helper
from scipy import ndimage
from skimage import feature
from os.path import join, basename
from scipy.stats import norm
import matplotlib.pyplot as plt
import pickle

Coordinates = collections.namedtuple('Coordinates', ['x', 'y', 'z'])


class Grid(utilities.Grid):
    """
    A class to extend a `ccdc.utilities.Grid` this provides grid methods required in the Fragment Hotspot Maps algorithm
    """
    def grid_value_by_coordinates(self, threshold=1):
        """
        returns a dictionary of grid point values by coordinates
        :param int threshold: the island threshold of the grid, only points over this value will be returned
        :return: dict, grid point values by coordinates
        """
        dic = {}
        nx, ny, nz = self.nsteps
        for i in range(nx):
            for j in range(ny):
                for k in range(nz):
                    if self.value(i, j, k) > threshold:
                        try:
                            dic[self.value(i, j, k)].append(self.indices_to_point(i, j, k))
                        except KeyError:
                            dic.update({self.value(i, j, k): [self.indices_to_point(i, j, k)]})

        return dic

    @staticmethod
    def _tolerance_range(value, tolerance, min, max):
        """
        private method

        for a given value and tolerance, the method checks that the tolerance range for that value is within the grid
        boundaries
        :param int value: an "indice" value either (i or j or k)
        :param int min: the minimum grid boundary (always = 0)
        :param int max: the maximum grid boundary (always = n step)
        :return: range,
        """
        low = value - tolerance
        if low < 0:
            low = 0

        high = value + tolerance
        if high > max:
            high = max
        return range(low, high)

    def get_near_scores(self, coordinate, tolerance=3):
        """
        for a given grid point, return a list of values within a search tolerance
        :param tup coordinate: coordinate of a point within the grid
        :param int tolerance: search distance, in grid steps
        :return:
        """
        i, j, k = self.point_to_indices(coordinate)
        ri = self._tolerance_range(i, tolerance, 0, self.nsteps[0])
        rj = self._tolerance_range(j, tolerance, 0, self.nsteps[1])
        rk = self._tolerance_range(k, tolerance, 0, self.nsteps[2])
        return [self.value(a, b, c) for a in ri for b in rj for c in rk if self.value(a,b,c) > 0]

    def grid_values(self, threshold=0):
        """
        generates a numpy array with all values over a given threshold (default = 0)
        :param int threshold: values over this value
        :return:
        """
        array = self.get_array()
        masked_array = np.ma.masked_less_equal(array, threshold)
        return masked_array.compressed()

    def grid_score(self, threshold=0, percentile=75):
        """
        for a given grid, the xth percentile of values above a given threshold is returned
        :param int threshold: values over this value
        :param int percentile: value at this percentile

        :return:
        """
        array = self.get_array()
        masked_array = np.ma.masked_less_equal(array, threshold)
        values = masked_array.compressed()

        if len(values) == 0:
            return 0
        else:
            return np.percentile(values, percentile)

    def indices_to_point(self, i, j, k):
        """
        return x,y,z coordinate for a given grid index
        :param int i: indice (x-axis)
        :param int j: indice (y-axis)
        :param int k: indice (z-axis)
        :return: tup, float(x), float(y), float(z)
        """
        ox, oy, oz, = self.bounding_box[0]
        gs = self.spacing
        return ox + float(i) * gs, oy + float(j) * gs, oz + gs * float(k)

    def point_to_indices(self, p):
        """
        return the nearest grid index for a given point
        :param tup p: (float(x), float(y), float(z)), a coordinate on a 3D grid
        :return: int(x), int(y), int(z)
        """

        gs = self.spacing
        rx, ry, rz = [round(i / gs) for i in p]
        ox, oy, oz = [round(i / gs) for i in self.bounding_box[0]]
        return int(rx - ox), int(ry - oy), int(rz - oz)

    def gaussian(self, sigma=0.2):
        """
        gaussian smoothing function, method of reducing noise in output
        :param float sigma: degree of smoothing
        :return:
        """
        s = (sigma, sigma, sigma, 0)
        nx, ny, nz = self.nsteps
        scores = np.zeros((nx, ny, nz, 1))
        for i in range(nx):
            for j in range(ny):
                for k in range(nz):
                    scores[i, j, k, 0] += self.value(i, j, k)
        smoothed = ndimage.filters.gaussian_filter(scores, sigma=s)
        grid = self.copy_and_clear()
        for i in range(nx):
            for j in range(ny):
                for k in range(nz):
                    grid.set_value(i, j, k, smoothed[i, j, k, 0])
        return grid

    def contains_point(self, point, threshold=0, tolerance=0):
        """
        determines whether a set of coordinates are within a grids boundary
        :param tup point: (float(x), float(y), float(z))
        :param int threshold: values above this value
        :param float tolerance: radius of search
        :return:
        """
        mini = self.bounding_box[0]
        maxi = self.bounding_box[1]
        if self.value_at_point(point) >= threshold:
            return all([mini.x - tolerance < point[0] < maxi.x + tolerance,
                        mini.y - tolerance < point[1] < maxi.y + tolerance,
                        mini.z - tolerance < point[2] < maxi.z + tolerance])
        else:
            return False

    def get_array(self):
        """
        convert grid object to np.array
        :return: `numpy.array`, array with shape (nsteps) and each element corresponding to value at that indice
        """
        nx, ny, nz = self.nsteps
        array = np.zeros((nx, ny, nz))
        for i in range(nx):
            for j in range(ny):
                for k in range(nz):
                    array[i, j, k] += self.value(i, j, k)
        return array

    def restricted_volume(self, volume=75):
        """
        returns a grid with of a defined volume
        :param float volume: desired volume in Angstroms ^ 3
        :return: `hotspots.grid_extension.Grid`
        """
        grid = self.copy_and_clear()
        max_points = int(float(volume) / 0.125)
        nx, ny, nz = self.nsteps
        rank_dict = {}

        for i in range(nx):
            for j in range(ny):
                for k in range(nz):
                    value = float(self.value(i, j, k))
                    if value in rank_dict:
                        rank_dict[value].append((i, j, k))
                    else:
                        rank_dict.update({value: [(i, j, k)]})

        top_points = sorted((float(x) for x, y in rank_dict.iteritems()), reverse=True)
        indices = [pts for key in top_points for pts in rank_dict[key]]

        for i in indices[:max_points]:
            grid.set_value(i[0], i[1], i[2], self.value(i[0], i[1], i[2]))

        return Grid.super_grid(1, *grid.islands(threshold=1))

    def centroid(self):
        """
        returns centre of the grid's bounding box
        :param self:
        :return:
        """
        return ((self.bounding_box[0][0] + self.bounding_box[1][0])/2,
                (self.bounding_box[0][1] + self.bounding_box[1][1])/2,
                (self.bounding_box[0][2] + self.bounding_box[1][2])/2
                )

    def deduplicate(self, major, threshold=12, tolerance=2):
        """
        method to deduplicate two grids, used for charged-polar deduplication
        :param `ccdc.utilities.Grid` major: overriding grid
        :param int threshold: values above this value
        :param int tolerance: search radius for determining feature overlap
        :return:
        """
        if self.bounding_box[0] != major.bounding_box[0] or self.bounding_box[1] != major.bounding_box[1]:
            self = major.common_boundaries(self)

        all_islands = set([jsland for jsland in self.islands(threshold=threshold)])
        bin_islands = set([jsland for jsland in all_islands
                           for island in major.islands(threshold=threshold)
                           if jsland.contains_point(island.centroid(), tolerance=tolerance)
                           or jsland.count_grid() <= 8
                           or Helper.get_distance(jsland.centroid(), island.centroid()) < 4])

        retained_jslands = list(all_islands - bin_islands)

        if len(retained_jslands) == 0:
            blank = major.copy_and_clear()
            return blank
        else:
            temp = Grid.super_grid(0, *retained_jslands)
            blank = self.copy_and_clear()
            return blank.common_boundaries(temp)

    def copy_and_clear(self):
        """
        make a new empty grid, with the same dimensions as the old
        :return: `hotspots.grid_extension.Grid`
        """
        g = self.copy()
        g *= 0
        return g

    def common_boundaries(self, grid):
        """
        expands supplied grid to the size of self (supplied grid should be smaller than self)
        :param grid:
        :return:
        """
        reference = Grid.super_grid(0, self, grid)
        blank = reference.copy_and_clear()
        return Grid.super_grid(0, blank, grid)

    def multi_max_mask(self, grids):
        """
        for a self grid and collection of grids supplied, for each grid point, if the maximum grid value across all the
        grids belongs to the self grid, the value is assigned to a fresh grid.
        :param list grids: grids to be compared to self
        :return: `ccdc.utilities.Grid`
        """
        max_grids = [self > g for g in grids]
        blank = -self.copy_and_clear()
        return reduce(operator.__and__, max_grids, blank)

    def get_best_island(self, threshold, mode="count", peak=None):
        """
        returns the best grid island. Mode: "count" or "score"
        :param int threshold: island threshold
        :param str mode:
                       -"count" : returns island with most grid points above threshold
                       -"score" : returns island with the largest sum of all grid points over threshold
        :param tup peak: (float(x), float(y), float(z)) coordinates of peak in grid
        :return: `ccdc.utilities.Grid`, grid containing the best island
        """
        islands = self.islands(threshold)
        if len(islands) == 0:
            return None

        else:
            island_by_rank = {}
            if mode == "count":
                for island in islands:
                    if peak:
                        if island.contains_point(peak):
                            g = (island > threshold) * island
                            rank = g.count_grid()
                            island_by_rank.update({rank: island})
                        else:
                            continue
                    else:
                        g = (island > threshold) * island
                        rank = g.count_grid()
                        island_by_rank.update({rank: island})

            elif mode == "score":
                for island in islands:
                    if peak:
                        if island.contains_point(peak):
                            nx, ny, nz = island.nsteps
                            island_points = [island.value(i, j, k)
                                             for i in range(nx) for j in range(ny) for k in range(nz)
                                             if island.value(i, j, k) >= threshold]
                            rank = sum(island_points)
                            island_by_rank.update({rank: island})
                        else:
                            continue
                    else:
                        nx, ny, nz = island.nsteps
                        island_points = [island.value(i, j, k)
                                         for i in range(nx) for j in range(ny) for k in range(nz)
                                         if island.value(i, j, k) >= threshold]
                        rank = sum(island_points)
                        island_by_rank.update({rank: island})

            else:
                raise IOError("mode not supported")

            if len(island_by_rank) == 0:
                return None
            else:
                rank = sorted(island_by_rank.keys(), reverse=True)[0]
                return island_by_rank[rank]

    def minimal(self):
        """
        reduces grid size to the minimal dimensions
        :return: `ccdc.utilities.Grid`
        """
        try:
            return Grid.super_grid(1, *self.islands(threshold=1))
        except RuntimeError:
            return self

    def limit_island_size(self, npoints, threshold=10):
        """
        for a given grid, the number of points contained within the islands (threshold = x) is limited to npoints
        :param int npoints: maximum number of points in each island feature
        :param float threshold: values above this value
        :return:
        """
        g = (self > 10) * self
        all_islands = []
        for island in g.islands(threshold):
            if island.count_grid() > npoints:
                all_islands.append(island.top_points(npoints=npoints))
            else:
                all_islands.append(island)
        return Grid.super_grid(0, *all_islands)

    def top_points(self, npoints):
        """
        for a given grid, the top scoring n points are returned
        :param int npoints: number of points to be returned
        :return: `ccdc.ulilities.Grid`
        """
        pts = {}
        nx, ny, nz = self.nsteps

        for i in range(nx):
            for j in range(ny):
                for k in range(nz):
                    val = self.value(i, j, k)

                    if val in pts:
                        pts[val].append((i, j, k))
                    else:
                        pts.update({self.value(i, j, k): [(i, j, k)]})

        sorted_pts = sorted(pts.items(), key=lambda x: x[0], reverse=True)

        thres = self._get_threshold(sorted_pts, npoints=npoints)
        return (self > thres) * self

    def step_out_mask(self, nsteps=2):
        """
        add n step in all directions to Grid boundary, returns blank grid
        :param nsteps: the number of steps the grid is expanded
        :return: `hotspots.grid_extension.Grid`
        """
        origin = (self.bounding_box[0].x - (self.spacing * nsteps),
                  self.bounding_box[0].y - (self.spacing * nsteps),
                  self.bounding_box[0].z - (self.spacing * nsteps))

        far_corner = (self.bounding_box[1].x + (self.spacing * nsteps),
                      self.bounding_box[1].y + (self.spacing * nsteps),
                      self.bounding_box[1].z + (self.spacing * nsteps))

        return Grid(origin=origin,
                    far_corner=far_corner,
                    spacing=0.5,
                    default=0,
                    _grid=None)

    @staticmethod
    def _get_threshold(sorted_points, npoints):
        """
        private method
        returns the lower limit of the top n number of points in a grid
        :param dict sorted_points: {grid_value, (float(x), float(y), float(z))
        :param int npoints: number of grid points
        :return:
        """
        count = []
        for value, pts in sorted_points:
            count.extend(pts)
            if len(count) >= npoints:
                return value
            else:
                continue

    @staticmethod
    def from_array(fname):
        """
        creates a grid from array
        :param fname: path to pickled numpy array
        :return: `hotspots.grid_extension.Grid`
        """
        grid = Grid(origin=[-35.00, -42.00, 44.00],
                    far_corner=[59.00, 53.00, 54.00],
                    spacing=0.5,
                    default=0.0,
                    _grid=None)

        array = np.load(fname)
        indices = np.nonzero(array)
        values = array[indices]
        as_triads = zip(*indices)

        for (i, j, k), v in zip(as_triads, values):
            grid._grid.set_value(int(i), int(j), int(k), v)

        return grid

    @staticmethod
    def common_grid(grid_list, padding=1):
        """
        returns two grid with common boundaries
        :param list grid_list: list of `ccdc.utilities.Grid` instances
        :param padding: number of steps to add to the grid boundary
        :return:
        """
        sg = Grid.super_grid(padding, *grid_list)
        out_g = sg.copy()
        out_g *= 0
        out = [Grid.super_grid(padding, g, out_g) for g in grid_list]
        return out

    def inverse_single_grid(self, mask_dic):
        """
        from a single grid, uses mask_dic to return separated grids divide by interaction type
        :param mask_dic: dict
        :return: dict
        """
        sg = mask_dic["apolar"].common_boundaries(self)
        return {probe: grid * (grid & sg) for probe, grid in mask_dic.items()}

    @staticmethod
    def get_single_grid(grd_dict, mask=True):
        """
        Combines a dictionary of identifier (str): Grid (ccdc.utilties.Grid) to a single grid.
        Grid points of the single_grid are set to the maximum value at each point across all the input grids
        :param grd_dict:
        :param mask:
        :return:
        """
        mask_dic = {}

        if len(set([g.bounding_box for g in grd_dict.values()])) > 1:
            grd_dict = dict(zip(grd_dict.keys(), Grid.common_grid(grd_dict.values())))

        for probe, grid in grd_dict.items():
            other_grids = [grd_dict[p] for p in grd_dict.keys() if p != probe]
            mask_dic.update({probe: grid * grid.multi_max_mask(other_grids)})

        o = [g.bounding_box[0] for g in grd_dict.values()]
        origin = [min([coord.x for coord in o]),
                  min([coord.y for coord in o]),
                  min([coord.z for coord in o])
                  ]

        f = [g.bounding_box[1] for g in grd_dict.values()]
        far_corner = [max([coord.x for coord in f]),
                      max([coord.y for coord in f]),
                      max([coord.z for coord in f])
                      ]

        blank = Grid(origin=origin, far_corner=far_corner, spacing=0.5, default=0, _grid=None)

        if mask:
            return mask_dic, reduce(operator.add, mask_dic.values(), blank).minimal()
        else:
            return reduce(operator.add, mask_dic.values(), blank).minimal()

    def percentage_overlap(self, other):
        """
        find the percentage overlap of this grid with other.
        :param other: `hotspots.grid_extension.Grid`
        :return:`hotspots.grid_extension.Grid`
        """
        g, h = Grid.common_grid(grid_list=[self, other], padding=1)
        vol = (g > 0).count_grid()
        overlap = (g & h).count_grid()
        return (overlap / vol) * 100

    @staticmethod
    def from_molecule(mol, scaling=1):
        """
        generate a molecule mask where gp within the vdw radius of the molecule heavy atoms are set to 1.0
        :param mol: `ccdc.molecule.Molecule`
        :param padding: int
        :param scaling: float
        :return: `hotspots.grid_extension.Grid`
        """
        coords = [a.coordinates for a in mol.atoms]
        g = Grid.initalise_grid(coords=coords)
        for a in mol.heavy_atoms:
            g.set_sphere(point=a.coordinates,
                         radius=a.vdw_radius * scaling,
                         value=1,
                         scaling='None')
        return g > 0.1

    @staticmethod
    def initalise_grid(coords, padding=1):
        """
        creates a fresh grid using a list of coordinates to define the grid limits
        :param coords: list
        :param padding: int, extra spacing added to grid limits
        :return:
        """
        x = set()
        y = set()
        z = set()
        for c in coords:
            x.add(c.x)
            y.add(c.y)
            z.add(c.z)

        origin = Coordinates(x=round(min(x) - padding),
                             y=round(min(y) - padding),
                             z=round(min(z) - padding))

        far_corner = Coordinates(x=round(max(x) + padding),
                                 y=round(max(y) + padding),
                                 z=round(max(z) + padding))

        return Grid(origin=origin, far_corner=far_corner, spacing=0.5, default=0, _grid=None)

    @staticmethod
    def grow(inner, template, percentile= 80):
        """
        experimental
        Dilates grid to the points in the top percentile of the template
        :param template:
        :return:
        """
        expand = inner.max_value_of_neighbours().max_value_of_neighbours() > 0.1   # remove very small values
        outer = expand.__mul__(-inner) * template
        #test = expand*template
        threshold = np.percentile(a=outer.grid_values(threshold=1), q=int(percentile))

        return inner.__add__(outer > threshold)
        #return template * ( > threshold)

    def get_peaks(self, min_distance=6, cutoff=2):
        """
        find peak coordinates in grid
        :return:
        """
        peaks = feature.peak_local_max(self.get_array(),
                                       min_distance=min_distance,
                                       threshold_abs=cutoff)
        peak_by_value = {}
        for peak in peaks:
            val = self.value(int(peak[0]), int(peak[1]), int(peak[2]))
            if val > cutoff:
                if val in peak_by_value:
                    peak_by_value[val].append((peak[0], peak[1], peak[2]))
                else:
                    peak_by_value.update({val: [(peak[0], peak[1], peak[2])]})

        average_peaks = []
        for key in peak_by_value.keys():
            x = [point[0] for point in peak_by_value[key]]
            y = [point[1] for point in peak_by_value[key]]
            z = [point[2] for point in peak_by_value[key]]
            average_peaks.append(self.indices_to_point(int(sum(x) / len(x)),
                                                       int(sum(y) / len(y)),
                                                       int(sum(z) / len(z))
                                                       )
                                 )
        return average_peaks

    def value_at_coordinate(self, coordinates, tolerance=1, position=True):
        """
        Uses Grid.value() rather than Grid.value_at_point(). Chris Radoux reported speed issues.
        :param coordinates:
        :param tolerance:
        :return:
        """
        i, j, k = self.point_to_indices(coordinates)
        nx, ny, nz = self.nsteps
        scores = {}

        for di in range(-tolerance, +tolerance + 1):
            for dj in range(-tolerance, +tolerance + 1):
                for dk in range(-tolerance, +tolerance + 1):
                    if 0 < (i + di) < nx and 0 < (j + dj) < ny and 0 < (k + dk) < nz:
                        scores.update({self.value(i + di, j + dj, k + dk): (i + di, j + dj, k + dk)})

        if len(scores) > 0:
            score = sorted(scores.keys(), reverse=True)[0]

            if score < 0.1:
                score = 0
                point = (0, 0, 0)

            else:
                a, b, c = scores[score]
                point = self.indices_to_point(a, b, c)

        else:
            score = 0
            point = (0, 0, 0)

        if position:
            return score, point

        else:
            return score

utilities.Grid = Grid


class _GridEnsemble(object):
    """
    Experimental feature
    Class that handles a numpy array of tuples from Hotspot maps of a given probe type across multiple structures
    """

    def __init__(self):
        self.prot_name = None
        self.probe = None
        self.path_list = None
        self.tup_max_length = None
        self.results_array = None
        self.common_grid_origin = None
        self.common_grid_far_corner = None
        self.common_grid_nsteps = None
        self.nonzeros = None
        self.out_dir = None
        self.spacing = None

    ###### Functions that generate the ensemble and set attributes

    def _common_grids_from_grid_list(self, grid_list, fnames=None):
        """
        Gets the coordinates of a grid that can fit all other grids (to use as basis for self.results_array)
        :param grid_list: list of 'hotspots.grid_extension.Grid' objects
        :return: list of 'hotspots.grid_extension.Grid' objects
        """
        print("Making array grid {} {}".format(self.prot_name, self.probe))
        common_grids = Grid.common_grid(grid_list)

        assert (len(set([c.bounding_box for c in common_grids])) == 1), "Common grids don't have same frames"
        self.spacing = grid_list[0].spacing
        # assert (common_grids[0].spacing == 0.5), "Grid spacing not 0.5"
        self.tup_max_length = len(grid_list)
        self.common_grid_origin = common_grids[0].bounding_box[0]
        self.common_grid_far_corner = common_grids[0].bounding_box[1]
        self.common_grid_nsteps = common_grids[0].nsteps

        if fnames:
            self.path_list = fnames

        return common_grids

    def _common_grids_from_paths(self):
        """
        Gets the coordinates of a grid that can fit all other grids (to use as basis for self.results_array)
        :return: list of 'hotspots.grid_extension.Grid' objects
        """
        print('Making array grid')
        grid_list = [Grid.from_file(f) for f in self.path_list if self.probe in basename(f)]
        common_grids = self._common_grids_from_grid_list(grid_list)
        self.tup_max_length = len(grid_list)
        self.common_grid_origin = common_grids[0].bounding_box[0]
        self.common_grid_far_corner = common_grids[0].bounding_box[1]
        self.common_grid_nsteps = common_grids[0].nsteps


        return common_grids

    def _get_results_array(self, common_grids):
        """
        constructs the numpy array containing the list of tuples.
        Adjusts the coordinates based on the origins of the grids
        self.results_array = numpy array
        :return:
        """
        print('Making results array')

        results_array = np.zeros(self.common_grid_nsteps, dtype=tuple)
        # rec_spacing = 1 / self.spacing

        nx, ny, nz = self.common_grid_nsteps
        for c in common_grids:
            for x in range(nx):
                for y in range(ny):
                    for z in range(nz):
                        if c.value(x, y, z) != 0:
                            if isinstance(results_array[x][y][z], tuple):
                                results_array[x][y][z] += (c.value(x, y, z),)
                            else:
                                results_array[x][y][z] = (c.value(x, y, z),)

        self.results_array = results_array
        self.nonzeros = self.results_array.nonzero()

    def get_alternative_results_array(self, grid_list):
        """"
        Could be of use for larger ensembles.
        """
        # Find bounding box of smallest grid that fits all the grids
        dimensions = np.array([g.bounding_box for g in grid_list])

        self.tup_max_length = len(grid_list)
        self.common_grid_origin = tuple(np.min(dimensions[:, 0, :], axis=0))
        self.common_grid_far_corner = tuple(np.max(dimensions[:, 1, :], axis=0))

        origins = dimensions[:, 0, :]
        print(origins)
        far_corners = dimensions[:, 1, :]
        print(far_corners)
        if self.spacing:
            rec_spacing = 1 / self.spacing
        else:
            rec_spacing = 2

        origin_diff = (origins - self.common_grid_origin) * rec_spacing
        print(origin_diff)

        if (origins == origins[0]).all() and (far_corners == far_corners[0]).all():
            print("Grids have same dimensions")
            self.common_grid_nsteps = grid_list[0].nsteps
            self._get_results_array(grid_list)


        else:
            comm_grid = Grid(origin=self.common_grid_origin,
                             far_corner=self.common_grid_far_corner,
                             spacing=0.5,
                             _grid=None)
            self.common_grid_nsteps = comm_grid.nsteps

            results_array = np.zeros(self.common_grid_nsteps, dtype=tuple)
            assert len(origin_diff) == len(grid_list)

            for i in range(len(origin_diff)):
                g = grid_list[i]
                diff = origin_diff[i]
                d_x, d_y, d_z = diff
                nx, ny, nz = g.nsteps
                for x in range(nx):
                    for y in range(ny):
                        for z in range(nz):
                            if g.value(x, y, z) != 0:
                                if isinstance(results_array[x + int(d_x)][y + int(d_y)][z + int(d_z)], tuple):
                                    results_array[x + int(d_x)][y + int(d_y)][z + int(d_z)] += (g.value(x, y, z),)
                                else:
                                    results_array[x + int(d_x)][y + int(d_y)][z + int(d_z)] = (g.value(x, y, z),)

            self.results_array = results_array
            self.nonzeros = self.results_array.nonzero()

    #### Functions for analysing ensemble data #####

    def get_gridpoint_means(self):
        """
        For each tuple in the GridEnsemble, calculates the means of the tuple
        :return: list
        """
        means = [np.mean(val) for val in self.results_array[self.nonzeros]]
        return means

    def get_gridpoint_max(self):
        """
        For each tuple in the GridEnsemble, calculates the max of the tuple
        :return: list
        """
        maxes = [np.max(val) for val in self.results_array[self.nonzeros]]
        return maxes

    def get_gridpoint_ranges(self):
        """
        For each tuple in the GridEnsemble, returns the difference between max and mean
        :return: list
        """
        ranges = [np.max(val) - np.min(val) for val in self.results_array[self.nonzeros]]
        return ranges

    def get_gridpoint_means_spread(self):
        """
        For tuple in the GridEnsemble, calculates the difference in score between each point in the tuple and the mean of the tuple.
        :return: Python list
        """
        means_spread = []
        values = self.results_array[self.nonzeros]

        for v in values:
            num_zeros = self.tup_max_length - len(v)
            if num_zeros != 0:
                print('Number of zeros', num_zeros, 'out of ', self.tup_max_length)
            hist_arr = np.array(v)
            means_spread.extend(list(hist_arr - np.mean(hist_arr)))

        return means_spread

    # Functions for plotting histograms of analysed ensemble data ####
    def plot_gridpoint_spread(self):
        '''
        For each point in the 3D grid, plots the difference in score between each point in the tuple and the mean of the tuple.
        '''
        # Get the data:
        means_spread = self.get_gridpoint_means_spread()
        mean_arr = np.array(means_spread)
        (mu, sigma) = norm.fit(mean_arr)
        n, bins, patches = plt.hist(means_spread, bins=40, normed=1)
        # print(bins)
        # y_fit = np.random.normal(mu, sigma, np.shape(mean_arr))
        y = norm.pdf(bins, mu, sigma)
        plt.plot(bins, y, 'r--', linewidth=2)
        fit_bins = 0.5 * (bins[1:] + bins[:-1])
        y_fit = norm.pdf(fit_bins, mu, sigma)
        ss_res = np.sum((n - y_fit) ** 2)
        ss_tot = np.sum((n - np.mean(n)) ** 2)
        r2 = 1 - (ss_res / ss_tot)
        plt.title('Mu: {}, Sigma: {}, R^2 : {} '.format(round(mu, 2), round(sigma, 2), round(r2, 2)))
        hist_name = self.prot_name + '_{}_gridpoint_spread_fit'.format(self.probe)
        plt.savefig(join(self.out_dir, hist_name))
        # plt.show()
        plt.close()

    def plot_gridpoint_ranges(self):
        '''
        Plots the range of the tuple values for each point in the 3D grid
        :return: ranges = list of the tuple ranges at each point
        '''
        ranges = self.get_gridpoint_ranges()
        plt.hist(ranges, bins=40, normed=0)
        plt.title('Score ranges for {} {}'.format(self.prot_name, self.probe))
        hist_name = self.prot_name + '_{}_score_ranges'.format(self.probe)
        plt.savefig(join(self.out_dir, hist_name))
        # plt.show()
        plt.close()

    #### Functions that output analysed ensemble data as Grids ####

    def _make_grid(self, values):
        """
        Makes a grid to store output of ranges, max, means, etc
        Duplicating Grid.from_array
        :param values:
        :return:
        """
        grid = Grid(origin=self.common_grid_origin,
                    far_corner=self.common_grid_far_corner,
                    spacing=0.5,
                    default=0.0,
                    _grid=None)

        as_triads = zip(*self.nonzeros)
        for (i, j, k), v in zip(as_triads, values):
            grid._grid.set_value(int(i), int(j), int(k), v)

        return grid

    def output_grid(self, mode="max", save=True):
        """
        Ouputs ensemble information as ccdc.utilites.Grid
        :param mode: the operation to be done on each tuple in the grid. Can be "max", "mean", "ranges", or "frequency" (Length of tuple)
        :param save: bool, whether output should be written to disk
        :return: a class ccdc.utilities.Grid object
        """
        # Get the values
        if mode == "max":
            vals = self.get_gridpoint_max()
        elif mode == "mean":
            vals = self.get_gridpoint_means()
        elif mode == "ranges":
            vals = self.get_gridpoint_ranges()
        elif mode == "frequency":
            vals = [np.max(val) - np.min(val) for val in self.results_array[self.nonzeros]]
        else:
            print("Unrecognised mode: {}".format(mode))
            return

        # Fill the grid
        out_grid = self._make_grid(vals)

        if save:
            out_grid.write(join(self.out_dir, '{}_{}_{}.ccp4'.format(self.prot_name, mode, self.probe)))

        return out_grid

    ###### Saving and loading ensembles #########

    @staticmethod
    def load_GridEnsemble(filename):
        """
        Loads a pickled _GridEnsemble
        :param filename: str, full path to pickled grid
        :return: MegaGrid object
        """
        pickle_file = open(filename, 'rb')
        newGridEnsemble = pickle.load(pickle_file)
        return newGridEnsemble

    def pickle_GridEnsemble(self):
        """
        Saves _GridEnsembles as pickles.
        :return:
        """
        pickle.dump(self, open(join(self.out_dir, '{}_{}_GridEnsemble.p'.format(self.prot_name, self.probe)), 'wb'))

    ###### Functions to run full ensemble calculation and output grids (to integrate into main hotspots code ######

    def from_hotspot_maps(self, path_list, out_dir, prot_name, probe_name, mode="max"):
        """
        Creates a GridEnsemble from paths to Hotspot maps for a certain probe
        :param path_list: list of paths for the hotspot maps
        :param out_dir: path to where grids and histograms are saved
        :param prot_name: str
        :param probe_name: 'donor', 'acceptor', or 'apolar'
        :return:
        """
        print('In from_hotspot_maps')
        self.path_list = path_list
        self.out_dir = out_dir
        self.prot_name = prot_name
        self.probe = probe_name

        common_grids = self._common_grids_from_paths()
        # grid_list = [Grid.from_file(p) for p in self.path_list]
        # self.get_alternative_results_array(grid_list)
        self._get_results_array(common_grids)

        return self.output_grid(mode, save=False)

    def from_grid_list(self, grid_list, out_dir, prot_name, probe_name, mode="max"):
        """
        Creates a GridEnsemble from Hotspot maps for a certain probe
        :param grid_list:
        :param out_dir:
        :param prot_name:
        :param probe_name:
        :return:
        """
        self.out_dir = out_dir
        self.prot_name = prot_name
        self.probe = probe_name
        print("Making ensemble {} {}".format(self.prot_name, self.probe))

        common_grids = self._common_grids_from_grid_list(grid_list)
        self._get_results_array(common_grids)
        # self.get_alternative_results_array(grid_list)
        print(self.common_grid_origin, self.common_grid_far_corner)

        return self.output_grid(mode, save=False)




"""
The :mod:`hotspots.hs_pharmacophore` module contains classes for the
conversion of Grid objects to pharmacophore models.

The main class of the :mod:`hotspots.hs_pharmacophore` module is:

- :class:`hotspots.hs_pharmacophore.PharmacophoreModel`


A Pharmacophore Model can be generated directly from a :class:`hotspots.result.Result` :

>>> from hotspots.calculation import Runner

>>> r = Runner()
>>> result = r.from_pdb("1hcl")
>>> result.get_pharmacophore_model(identifier="MyFirstPharmacophore")

The Pharmacophore Model can be used in Pharmit or CrossMiner

>>> result.pharmacophore.write("example.cm")   # CrossMiner
>>> result.pharmacophore.write("example.json")    # Pharmit

More information about CrossMiner is available:
    - Korb O, Kuhn B, hert J, Taylor N, Cole J, Groom C, Stahl M "Interactive and Versatile Navigation of Structural Databases" J Med Chem, 2016, 59(9):4257, [DOI: 10.1021/acs.jmedchem.5b01756]

More information about Pharmit is available:
    - Jocelyn Sunseri, David Ryan Koes; Pharmit: interactive exploration of chemical space, Nucleic Acids Research, Volume 44, Issue W1, 8 July 2016, Pages W442-W448 [DIO: 10.1093/nar/gkw287]

"""
from __future__ import print_function
import csv
import json
import os
import re
import tempfile
from os.path import basename, splitext, join, dirname

import numpy as np
from ccdc import io
from ccdc.descriptors import GeometricDescriptors
from ccdc.molecule import Atom, Molecule
from ccdc.pharmacophore import Pharmacophore
from ccdc.protein import Protein

from hotspots.grid_extension import Grid, Coordinates
from hotspots.hs_utilities import Helper
from hotspots.template_strings import pymol_arrow, pymol_imports, crossminer_features, pymol_labels
from hotspots.pdb_python_api import Query, PDB, PDBResult

from rdkit import Chem, DataStructs
# from rdkit.Chem import MACCSkey, AllChem, Draw
from rdkit.Chem import MACCSkeys
from sklearn.cluster import KMeans
from sklearn.metrics import silhouette_score
from tqdm import tqdm


class PharmacophoreModel(Helper):
    """
    A class to handle a Pharmacophore Model

    :param `hotspots.hs_pharmacophore.PharmacophoreModel.Settings` settings: Pharmacophore Model settings
    :param str identifier: Model identifier
    :param list features: list of  :class:hotspots.hs_pharmacophore._PharmacophoreFeatures
    :param `ccdc.protein.Protein` protein: a protein
    :param dict dic: key = grid identifier(interaction type), value = :class:`ccdc.utilities.Grid`

    """

    class Settings():
        """
        settings available for adjustment

        :param float feature_boundary_cutoff: The map score cutoff used to generate islands
        :param float max_hbond_dist: Furthest acceptable distance for a hydrogen bonding partner (from polar feature)
        :param float radius: Sphere radius
        :param bool vector_on: Include interaction vector
        :param float transparency: Set transparency of sphere
        :param bool excluded_volume:  If True, the CrossMiner pharmacophore will contain excluded volume spheres
        :param float binding_site_radius: Radius of search for binding site calculation, used for excluded volume
        """

        def __init__(self, feature_boundary_cutoff=5, max_hbond_dist=5, radius=1.0, vector_on=False, transparency=0.6,
                     excluded_volume=True, binding_site_radius=12):
            self.feature_boundary_cutoff = feature_boundary_cutoff
            self.max_hbond_dist = max_hbond_dist
            self.radius = radius  # set more intelligently
            self.transparency = transparency
            self.excluded_volume = excluded_volume
            self.binding_site_radius = binding_site_radius

            if vector_on:
                self.vector_on = 1
            else:
                self.vector_on = 0

    def __init__(self, settings, identifier=None, features=None, protein=None, dic=None):
        self.identifier = identifier
        self._features = features

        self.fname = None
        self.projected_dict = {"True": ["donor", "acceptor"], "False": ["negative", "positive", "apolar"]}
        self.protein = protein

        if settings == None:
            self.settings = self.Settings()
        else:
            self.settings = settings

        self.dic = dic

    @property
    def features(self):
        """
        Interaction features of the Pharmacophore Model

        :return: list of :class:`hotspots.hs_pharmacophore._PharmacophoreFeature instance
        """
        return self._features

    def rank_features(self, max_features=4, feature_threshold=0, force_apolar=True):
        """
        orders features by score

        :param int max_features: maximum number of features returned
        :param float feature_threshold: only features above this value are considered
        :param force_apolar: ensures at least one point is apolar
        :return: list of features


        >>> from hotspots.hs_io import HotspotReader

        >>> result = HotspotReader("out.zip").read()
        >>> model = result.get_pharmacophore_model()
        >>> print(len(model.features))
        38
        >>> model.rank_features(max_features=5)
        >>> print(len(model.features))
        5

        """
        if force_apolar:
            apolar_dict = {feat.score_value: feat for feat in self.features if feat.feature_type == "apolar"}
            top_apolar = sorted(apolar_dict.items(), key=lambda x: x[0], reverse=True)[0]
            apolar = apolar_dict[top_apolar[0]]

            score_dic = {feat.score_value: feat for feat in self.features if feat.feature_type != "apolar"}
            sorted_scores = sorted(score_dic.items(), key=lambda x: x[0], reverse=True)
            max_features -= 1
            if max_features > len(self.features):
                ordered_features = [feat for score, feat in sorted_scores if score > feature_threshold]
            else:
                ordered_features = [feat for score, feat in sorted_scores if score > feature_threshold][:max_features]

            ordered_features.append(apolar)

        else:
            score_dic = {feat.score_value: feat for feat in self.features}
            sorted_scores = sorted(score_dic.items(), key=lambda x: x[0], reverse=True)
            if max_features > len(self.features):
                ordered_features = [feat for score, feat in sorted_scores if score > feature_threshold]
            else:
                ordered_features = [feat for score, feat in sorted_scores if score > feature_threshold][:max_features]

        self._features = ordered_features

    def _get_pymol_pharmacophore(self, lfile):
        """
        create the pymol visualisation for a pharmacophore
        """
        pymol_out = """
cluster_dict = {{"{0}":[], "{0}_arrows":[]}}
""".format(self.identifier)
        sphere_dict = {'acceptor': '[COLOR, 1.00, 0.00, 0.00]',
                       'donor': '[COLOR, 0.00, 0.00, 1.00]',
                       'apolar': '[COLOR, 1.00, 1.000, 0.000]',
                       'surface': '[COLOR, 0.5, 0.5, 0.5]',
                       'positive': '[COLOR, 0.0, 1.0, 1.0]',
                       'negative': '[COLOR, 0.6, 0.1, 0.6]'
                       }
        colour_dict = {'acceptor': 'red blue',
                       'donor': 'blue red',
                       'apolar': 'yellow'}
        i = 0

        for feature in self.features:
            if feature.feature_type in self.projected_dict["True"] and feature.projected_coordinates is not None:
                i += 1
                arrow = 'cluster_dict["{7}_arrows"] += cgo_arrow([{0},{1},{2}], [{3},{4},{5}], color="{6}", name="Arrows_{7}_{8}")\n' \
                    .format(feature.feature_coordinates.x,
                            feature.feature_coordinates.y,
                            feature.feature_coordinates.z,
                            feature.projected_coordinates.x,
                            feature.projected_coordinates.y,
                            feature.projected_coordinates.z,
                            colour_dict[feature.feature_type],
                            self.identifier,
                            str(i))
            else:
                arrow = ''

            sphere = '{0} + [ALPHA, {1}] + [SPHERE, float({2}), float({3}), float({4}), float({5})]\n' \
                .format(sphere_dict[feature.feature_type],
                        feature.settings.transparency,
                        feature.feature_coordinates.x,
                        feature.feature_coordinates.y,
                        feature.feature_coordinates.z,
                        feature.settings.radius)

            pymol_out += '\ncluster_dict["{0}"] += {1}'.format(self.identifier, sphere)
            pymol_out += '\n{}'.format(arrow)

        pymol_out += '\ncmd.load_cgo(cluster_dict["{0}"], "Features_{0}", 1)' \
                     '\ncmd.load_cgo(cluster_dict["{0}_arrows"], "Arrows_{0}")'.format(self.identifier)
        pymol_out += '\ncmd.set("transparency", 0.2,"Features_{0}")' \
                     '\ncmd.group("Pharmacophore_{0}", members="Features_{0}")' \
                     '\ncmd.group("Pharmacophore_{0}", members="Arrows_{0}")\n'.format(self.identifier)

        pymol_out += pymol_labels(fname=lfile,
                                  objname="label_threshold_{}".format(self.identifier))

        pymol_out += """\ncmd.group('Pharmacophore_{0}', members= 'label_threshold_{0}')\n""".format(self.identifier)

        return pymol_out

    def _as_grid(self, feature_type=None, tolerance=2):
        """
        returns _features as grid
        """
        if feature_type == None:
            filtered_features = self._features

        else:
            filtered_features = [feat for feat in self._features if feat.feature_type == feature_type]

        x = [feat.feature_coordinates.x for feat in filtered_features]
        y = [feat.feature_coordinates.y for feat in filtered_features]
        z = [feat.feature_coordinates.z for feat in filtered_features]

        origin = [min(x) - tolerance, min(y) - tolerance, min(z) - tolerance]
        far_corner = [max(x) + tolerance, max(y) + tolerance, max(z) + tolerance]
        grd = Grid(origin=origin, far_corner=far_corner, spacing=0.5, default=0, _grid=None)

        for feat in filtered_features:
            grd.set_sphere(point=feat.feature_coordinates,radius=self.settings.radius, value=1,scaling='None')
        return grd

    def _get_binding_site_residues(self):
        """
        return a protein binding site
        """
        centroid = [feat.feature_coordinates for feat in self._features if feat.feature_type == "apolar"][0]
        prot = self.protein

        bs = prot.BindingSiteFromPoint(protein=self.protein,
                                       origin=centroid,
                                       distance=self.settings.binding_site_radius)

        bs_residues = [str(r.identifier) for r in bs.residues]
        protein_residues = [str(p.identifier) for p in prot.residues]
        deletes = list(set(protein_residues) - set(bs_residues))
        for delete in deletes:
            prot.remove_residue(delete)

        return prot

    def _get_crossminer_pharmacophore(self):
        """
        convert a PharmacophoreModel into a crossminer pharmacophore
        """
        # TODO: UPDATE WITH CHARGED FEATURES
        supported_features = {"acceptor_projected": "acceptor",
                              "donor_projected": "donor",
                              "ring": "apolar"}
        try:
            Pharmacophore.read_feature_definitions()
        except:
            raise ImportError("Crossminer is only available to CSD-Discovery")

        feature_definitions = {supported_features[fd.identifier]: fd for fd in Pharmacophore.feature_definitions.values()
                               if fd.identifier in supported_features.keys()}

        model_features = []
        for feat in self._features:
            if feat.feature_type == "negative" or feat.feature_type == "positive":
                print("Charged feature not currently supported in CrossMiner: Its on the TODO list")

            else:
                sphere = GeometricDescriptors.Sphere(feat.feature_coordinates, self.settings.radius)

                if feat.projected_coordinates:
                    projected = GeometricDescriptors.Sphere(feat.projected_coordinates, self.settings.radius)
                    p = Pharmacophore.Feature(feature_definitions[feat.feature_type], *[sphere, projected])

                else:
                    p = Pharmacophore.Feature(feature_definitions[feat.feature_type], sphere)

                model_features.append(p)

        if self.settings.excluded_volume:
            if not self.protein:
                print("Pharmacophore Model must have protein to calculate excluded volume")
            else:
                bs = self._get_binding_site_residues()

                for residue in bs.residues:
                    mol = None
                    mol = Molecule(identifier="temp_residue")

                    # for a in residue.backbone_atoms:
                    #     ev = Pharmacophore.ExcludedVolume(GeometricDescriptors.Sphere(a.coordinates, 2))
                    #     model_features.append(ev)
                    for a in residue.backbone_atoms:
                        mol.add_atom(a)

                    centre = mol.centre_of_geometry()
                    ev = Pharmacophore.ExcludedVolume(GeometricDescriptors.Sphere(centre, 2))
                    model_features.append(ev)

        return Pharmacophore.Query(model_features)

    @staticmethod
    def _run_query(accession_id):
        """
        search the PDB for entries which have the same UniProt code, contain ligands and are
        within a resolution cutoff

        :param str accession_id: the accession ID for the target
        :return list: list of :class:`pdb_python_api.PDBResult` instances
        """
        # create query
        q = Query()
        q.add_term(query_type="UpAccessionIdQuery",
                   query_parameters={'accessionIdList': '{}'.format(accession_id)})

        q.add_term(query_type="NoLigandQuery",
                   query_parameters={'haveLigands': 'yes'},
                   conjunction='and')

        q.add_term(query_type="ResolutionQuery",
                   query_parameters={'refine.ls_d_res_high.comparator': 'between',
                                     'refine.ls_d_res_high.min': '0',
                                     'refine.ls_d_res_high.max': '2.5'},
                   conjunction='and')

        return PDB.search(q.query)

    @staticmethod
    def _get_ligands(results):
        """
        smiles to RDKit molecule

        :param list: list of :class:`pdb_python_api.PDBResult` instances
        :return: RDKit molecules
        """
        ligs = []
        uniques = []
        for entry in results:
            for l in entry.filtered_ligands:
                try:
                    l.rdmol = Chem.MolFromSmiles(l.smiles)
                    l.rdmol.SetProp("_Name", str(entry.identifier + "/" + l.chemical_id))
                    l.fingerprint = MACCSkeys.GenMACCSKeys(l.rdmol)
                    if l.chemical_id not in uniques:
                        ligs.append(l)
                        uniques.append(l.chemical_id)
                except AttributeError:
                    continue
        return ligs

    @staticmethod
    def _cluster_ligands(n, ligands):
        """
        !!! OUT OF DATE !!!
        :return:
        """
        cluster_dict = {a: [] for a in range(int(n))}
        num = len(ligands)
        sim = np.zeros((num, num))
        for i in range(num):
            for j in range(num):
                sim[i, j] = DataStructs.FingerprintSimilarity(ligands[i].fingerprint,
                                                              ligands[j].fingerprint)

        kmeans_model = KMeans(n_clusters=n, random_state=1).fit(sim)
        labels = kmeans_model.labels_
        s = silhouette_score(sim, labels, metric='euclidean')

        for i, ident in enumerate(kmeans_model.labels_):
            ligands[i].cluster_id = ident
            cluster_dict[int(ident)].append(ligands[i])

        return cluster_dict, s

    @staticmethod
    def _align_proteins(reference, reference_chain, targets):
        """
        align proteins by chain

        :param `ccdc.protein.Protein` reference: align to me
        :param str reference_chain: align to this chain
        :param list targets: list of `ccdc.protein.Protein`
        :return tup: list(:class:`ccdc.protein.Protein`) and list (:classa`ccdc.molecule.Molecule`)
        """
        print("Aligning proteins to {}, chain {}...".format(reference.identifier, reference_chain))
        aligned_prots = []
        aligned_ligands = []

        reference = Protein.from_file(reference.fname)
        reference.add_hydrogens()

        for t in tqdm(targets):
            prot = Protein.from_file(t.fname)
            prot.detect_ligand_bonds()
            prot.add_hydrogens()
            for l in prot.ligands:
                if str(t.clustered_ligand) == str(l.identifier.split(":")[1][0:3]):
                    try:
                        bs = Protein.BindingSiteFromMolecule(protein=prot,
                                                             molecule=l,
                                                             distance=6)
                        chain = bs.residues[0].identifier.split(":")[0]
                    except:
                        break
                    break

                else:
                    continue
            if not chain:
                print("\n        {} failed! No chain detected".format(t.identifier))
                break
            try:
                binding_site_superposition = Protein.ChainSuperposition()
                (bs_rmsd, bs_transformation) = binding_site_superposition.superpose(reference[reference_chain],
                                                                                    prot[chain])
                aligned_prots.append(prot)
                for lig in prot.ligands:
                    if str(t.clustered_ligand) == str(lig.identifier.split(":")[1][0:3]):
                        if chain == str(lig.identifier.split(":")[0]):
                            aligned_ligands.append(lig)
            except IndexError:
                print("\n        {} failed!".format(t.identifier))
                continue

        return aligned_prots, aligned_ligands

    def write(self, fname):
        """
        writes out pharmacophore. Supported formats:

        - ".cm" (*CrossMiner*),
        - ".json" (`Pharmit <http://pharmit.csb.pitt.edu/search.html/>`_),
        - ".py" (*PyMOL*),
        - ".csv",
        - ".mol2"

        :param str fname: path to output file
        """
        extension = splitext(fname)[1]

        if extension == ".cm":

            print("WARNING! Charged features not currently supported in CrossMiner!")
            pharmacophore = self._get_crossminer_pharmacophore()
            pharmacophore.write(fname)

        elif extension == ".csv":
            with open(fname, "wb") as csv_file:
                csv_writer = csv.writer(csv_file, delimiter=",")
                line = 'Identifier, Feature_type, x, y, z, score, ' \
                       'projected_x, projected_y, projected_z, ' \
                       'vector_x, vector_y, vector_z'

                for feature in self._features:
                    line += "{0},{1},{2},{3},{4},{5}".format(self.identifier,
                                                            feature.feature_type,
                                                            feature.feature_coordinates.x,
                                                            feature.feature_coordinates.y,
                                                            feature.feature_coordinates.z,
                                                            feature.score_value
                                                            )
                    if feature.projected_coordinates:
                        line += ",{0},{1},{2}".format(feature.projected_coordinates.x,
                                                      feature.projected_coordinates.y,
                                                      feature.projected_coordinates.z)
                    else:
                        line += ",0,0,0"

                    if feature.vector:
                        line += ",{0},{1},{2}".format(feature.vector.x,
                                                      feature.vector.y,
                                                      feature.vector.z)
                    else:
                        line += ",0,0,0"

                    l = line.split(",")
                    csv_writer.writerow(l)

        elif extension == ".py":
            with open(fname, "wb") as pymol_file:
                lfile = "label_threshold_{}.mol2".format(self.identifier)

                pymol_out = pymol_imports()
                pymol_out += pymol_arrow()
                lines = self._get_pymol_pharmacophore(lfile)
                pymol_out += lines
                pymol_file.write(pymol_out)

            label = self.get_label(self)
            with io.MoleculeWriter(join(dirname(fname),lfile)) as writer:
                writer.write(label)

        elif extension == ".json":
            with open(fname, "w") as pharmit_file:
                pts = []
                interaction_dic = {'apolar': 'Hydrophobic',
                                   'donor': 'HydrogenDonor',
                                   'acceptor': 'HydrogenAcceptor',
                                   'negative': 'NegativeIon',
                                   'positive': 'PositiveIon'
                                   }

                for feat in self._features:
                    if feat.vector:
                        point = {"name": interaction_dic[feat.feature_type],
                                 "hasvec": True,
                                 "x": feat.feature_coordinates.x,
                                 "y": feat.feature_coordinates.y,
                                 "z": feat.feature_coordinates.z,
                                 "radius": feat.settings.radius,
                                 "enabled": True,
                                 "vector_on": feat.settings.vector_on,
                                 "svector": {"x": feat.vector.x,
                                             "y": feat.vector.y,
                                             "z": feat.vector.z},
                                 "minsize": "",
                                 "maxsize": "",
                                 "selected": False
                                 }
                    else:
                        point = {"name": interaction_dic[feat.feature_type],
                                 "hasvec": False,
                                 "x": feat.feature_coordinates.x,
                                 "y": feat.feature_coordinates.y,
                                 "z": feat.feature_coordinates.z,
                                 "radius": feat.settings.radius,
                                 "enabled": True,
                                 "vector_on": feat.settings.vector_on,
                                 "svector": {"x": 0,
                                             "y": 0,
                                             "z": 0},
                                 "minsize": "",
                                 "maxsize": "",
                                 "selected": False
                                 }
                    pts.append(point)
                pharmit_file.write(json.dumps({"points": pts}))

        elif extension == ".mol2":
            mol = Molecule(identifier = "pharmacophore_model")
            atom_dic = {"apolar": 'C',
                        "donor": 'N',
                        "acceptor": 'O',
                        "negative": 'S',
                        "positve": 'H'}

            pseudo_atms = [Atom(atomic_symbol=atom_dic[feat.feature_type],
                                atomic_number=14,
                                coordinates=feat.feature_coordinates,
                                label=str(feat.score_value))
                           for feat in self.features]

            for a in  pseudo_atms:
                mol.add_atom(a)

            with io.MoleculeWriter(fname) as w:
                w.write(mol)

        elif extension == ".grd":
            g = self._as_grid()
            g.write(fname)

        else:
            raise TypeError("""""{}" output file type is not currently supported.""".format(extension))

    @staticmethod
    def from_hotspot(result, identifier="id_01", threshold=5, settings=None):
        """
        creates a pharmacophore model from a Fragment Hotspot Map result

        (included for completeness, equivalent to `hotspots.result.Result.get_pharmacophore()`)

        :param `hotspots.result.Result` result: a Fragment Hotspot Maps result (or equivalent)
        :param str identifier: Pharmacophore Model identifier
        :param float threshold: values above this value
        :param `hotspots.hs_pharmacophore.PharmacophoreModel.Settings` settings: settings

        :return: :class:`hotspots.hs_pharmacophore.PharmacophoreModel`

        >>> from hotspots.calculation import Runner
        >>> from hotspots.hs_pharmacophore import PharmacophoreModel

        >>> r = Runner()
        >>> result = r.from_pdb("1hcl")
        >>> model = PharmacophoreModel(result, identifier="pharmacophore")

        """

        if not settings:
            settings = PharmacophoreModel.Settings()

        feature_list = [_PharmacophoreFeature.from_hotspot(island, probe, result.protein, settings)
                        for probe, g in result.super_grids.items()
                        for island in g.islands(threshold) if island.count_grid() >= 5]

        return PharmacophoreModel(settings,
                                  identifier=identifier,
                                  features=feature_list,
                                  protein=result.protein)

    @staticmethod
    def _from_file(fname, protein=None, identifier=None, settings=None):
        """
        creates a pharmacophore model from file (only .cm supported)

        :param str fname: path to CrossMiner file
        :param `ccdc.protein.Protein` protein: protein
        :param str identifier: identifier for the Pharmacophore Model
        :param `hotspots.hs_pharmacophore.PharmacophoreModel.Settings` settings: settings

        :return: :class:`hotspots.hs_pharmacophore.PharmacophoreModel`

        """
        if not settings:
            settings = PharmacophoreModel.Settings()

        if identifier is None:
            identifier = basename(fname).split(".")[0]

        with open(fname) as f:
            file = f.read().split("FEATURE_LIBRARY_END")[1]
            lines = [l for l in file.split("""\n\n""") if l != ""]
            feature_list = [f for f in [_PharmacophoreFeature.from_crossminer(feature) for feature in lines] if f != None]

        return PharmacophoreModel(settings,
                                  identifier=identifier,
                                  features=feature_list,
                                  protein=protein)

    @staticmethod
    def from_ligands(ligands, identifier, protein=None, settings=None):
        """
        creates a Pharmacophore Model from a collection of overlaid ligands

        :param `ccdc,molecule.Molecule` ligands: ligands from which the Model is created
        :param str identifier: identifier for the Pharmacophore Model
        :param `ccdc.protein.Protein` protein: target system that the model has been created for
        :param `hotspots.hs_pharmacophore.PharmacophoreModel.Settings` settings: Pharmacophore Model settings

        :return: :class:`hotspots.hs_pharmacophore.PharmacophoreModel`


        >>> from ccdc.io import MoleculeReader
        >>> from hotspots.hs_pharmacophore import PharmacophoreModel

        >>> mols = MoleculeReader("ligand_overlay_model.mol2")
        >>> model = PharmacophoreModel.from_ligands(mols, "ligand_overlay_pharmacophore")
        >>> # write to .json and search in pharmit
        >>> model.write("model.json")

        """
        cm_dic = crossminer_features()
        blank_grd = Grid.initalise_grid([a.coordinates for l in ligands for a in l.atoms])
        feature_dic = {"apolar": blank_grd.copy(),
                       "acceptor": blank_grd.copy(),
                       "donor": blank_grd.copy()}

        if not settings:
            settings = PharmacophoreModel.Settings()

        if isinstance(ligands[0], Molecule):
            temp = tempfile.mkdtemp()

            with io.MoleculeWriter(join(temp, "ligs.mol2")) as w:
                for l in ligands:
                    w.write(l)
            ligands = list(io.CrystalReader(join(temp, "ligs.mol2")))

        try:
            Pharmacophore.read_feature_definitions()
        except:
            raise ImportError("Crossminer is only available to CSD-Discovery")

        feature_definitions = [fd for fd in Pharmacophore.feature_definitions.values() if
                               fd.identifier != 'exit_vector' and
                               fd.identifier != 'heavy_atom' and
                               fd.identifier != 'hydrophobe' and
                               fd.identifier != 'fluorine' and
                               fd.identifier != 'bromine' and
                               fd.identifier != 'chlorine' and
                               fd.identifier != 'iodine' and
                               fd.identifier != 'halogen']

        for fd in feature_definitions:
            detected = [fd.detect_features(ligand) for ligand in ligands]
            all_feats = [f for l in detected for f in l]

            if not all_feats:
                    continue

            for f in all_feats:
                feature_dic[cm_dic[fd.identifier]].set_sphere(f.spheres[0].centre, f.spheres[0].radius, 1)

        features = []
        for feat, feature_grd in feature_dic.items():
            peaks = feature_grd.get_peaks(min_distance=4, cutoff=1)
            for p in peaks:
                coords = Coordinates(p[0], p[1], p[2])
                projected_coordinates = None
                if feat == "donor" or feat == "acceptor":
                    if protein:
                        projected_coordinates = _PharmacophoreFeature.get_projected_coordinates(feat,
                                                                                                coords,
                                                                                                protein,
                                                                                                settings)
                features.append(_PharmacophoreFeature(projected=None,
                                                      feature_type=feat,
                                                      feature_coordinates=coords,
                                                      projected_coordinates=projected_coordinates,
                                                      score_value=feature_grd.value_at_coordinate(coords, position=False),
                                                      vector=None,
                                                      settings=settings
                                                      )
                                )

        return PharmacophoreModel(settings,
                                  identifier=identifier,
                                  features=features,
                                  protein=protein,
                                  dic=feature_dic)
    @staticmethod
    def _to_file(ligands, out_dir, fname="ligands.dat"):
        with open(os.path.join(out_dir, fname), "w") as f:
            for l in ligands:
                f.write("{},{},{}\n".format(l.structure_id, l.chemical_id, l.smiles))

    @staticmethod
    def from_pdb(pdb_code, chain, out_dir=None, representatives=None, identifier="LigandBasedPharmacophore"):
        """
        creates a Pharmacophore Model from a PDB code.

        This method is used for the creation of Ligand-Based pharmacophores. The PDB is searched for protein-ligand
        complexes of the same UniProt code as the input. These PDB's are align, the ligands are clustered and density
        of atom types a given point is assigned to a grid.

        :param str pdb_code: single PDB code from the target system
        :param str chain: chain of interest
        :param str out_dir: path to output directory
        :param representatives: path to .dat file containing previously clustered data (time saver)
        :param str identifier: identifier for the Pharmacophore Model

        :return: :class:`hotspots.hs_pharmacophore.PharmacophoreModel`


        >>> from hotspots.hs_pharmacophore import PharmacophoreModel
        >>> from hotspots.result import Results
        >>> from hotspots.hs_io import HotspotWriter
        >>> from ccdc.protein import Protein
        >>> from pdb_python_api import PDBResult


        >>> # get the PDB ligand-based Pharmacophore for CDK2
        >>> model = PharmacophoreModel.from_pdb("1hcl")

        >>> # the models grid data is stored as PharmacophoreModel.dic
        >>> # download the PDB file and create a Results
        >>> PDBResult("1hcl").download(<output_directory>)
        >>> result = Result(protein=Protein.from_file("<output_directory>/1hcl.pdb"), super_grids=model.dic)
        >>> with HotspotWriter("<output_directory>") as w:
        >>>     w.write(result)

        """
        temp = tempfile.mkdtemp()
        ref = PDBResult(pdb_code)
        ref.download(out_dir=temp, compressed=False)

        if representatives:
            print("Reading representative PDB codes ...")
            reps = []
            f = open(representatives, "r")
            entries = f.read().splitlines()

            for entry in entries:
                pdb_code, hetid, smiles = entry.split(",")
                reps.append((pdb_code, hetid))

        else:
            accession_id = PDBResult(pdb_code).protein.sub_units[0].accession_id
            results = PharmacophoreModel._run_query(accession_id)
            ligands = PharmacophoreModel._get_ligands(results)

            top = os.path.dirname(os.path.dirname(out_dir))
            PharmacophoreModel._to_file(ligands, top, fname="all_ligands.dat")

            k = int(round(len(ligands) / 5))
            if k < 2:
                k = 2
            cluster_dict, s = PharmacophoreModel._cluster_ligands(n=k, ligands=ligands)
            reps = [l[0] for l in cluster_dict.values() if len(l) != 0]

        targets = []

        for rep in reps:
            try:
                r = PDBResult(identifier=rep.structure_id)
                r.clustered_ligand = rep.chemical_id

            except:
                try:
                    r = PDBResult(identifier=rep[0])
                    r.clustered_ligand = rep[1]
                except:
                    raise AttributeError

            r.download(out_dir=temp, compressed=False)
            targets.append(r)

        prots, ligands = PharmacophoreModel._align_proteins(reference=ref,
                                                            reference_chain=chain,
                                                            targets=targets)

        try:
            with open(os.path.join(top, "silhouette.dat"), "w") as sfile:
                sfile.write("{}".format(s))
            PharmacophoreModel._to_file(reps, top, fname="representatives.dat")

        except:
            pass

        if out_dir:
            with io.MoleculeWriter(os.path.join(out_dir, "aligned_mols.mol2")) as w:
                for l in ligands:
                    w.write(l)

        return PharmacophoreModel.from_ligands(ligands=ligands,
                                               identifier=identifier)


class _PharmacophoreFeature(Helper):
    """
    A class to construct pharmacophoric models based upon fragment hotspot maps.

    :param projected:
    :param feature_type:
    :param feature_coordinates:
    :param projected_coordinates:
    :param score_value:
    :param vector:
    :param settings:

    """
    def __init__(self, projected, feature_type, feature_coordinates, projected_coordinates, score_value, vector, settings):
        self._projected = projected
        self._feature_type = feature_type
        self._feature_coordinates = feature_coordinates
        self._projected_coordinates = projected_coordinates
        self._score_value = score_value
        self._vector = vector

        self.settings = settings

        if feature_type == "donor" or "acceptor":
            self.settings.vector_on = 1

    @property
    def projected(self):
        return self._projected

    @property
    def feature_type(self):
        return self._feature_type

    @property
    def feature_coordinates(self):
        return self._feature_coordinates

    @property
    def projected_coordinates(self):
        return self._projected_coordinates

    @property
    def score_value(self):
        return self._score_value

    @property
    def vector(self):
        return self._vector


    @staticmethod
    def from_hotspot(grid, probe, protein, settings):
        """
        create a feature from a hotspot island

        :param `ccdc.utilities.Grid`grid: input grid
        :param str probe: probe type identifier
        :param `ccdc.protein.Protein` protein: target protein
        :param `hotspots.hs_pharmacophore.PharmacophoreModel.Settings` settings: settings
        :return: :class:`hotspots.hs_pharmacophore._PharmacophoreFeature`
        """
        feature_type = probe
        if probe == "apolar":
            score, feature_coordinates = _PharmacophoreFeature.get_centroid(grid)
            projected = False
            projected_coordinates = None
            vector = None

        else:
            vector = None
            projected_coordinates = None
            score, feature_coordinates = _PharmacophoreFeature.get_maxima(grid)
            if probe == "donor" or probe == "acceptor":

                projected = True
                if protein:
                    projected_coordinates = _PharmacophoreFeature.get_projected_coordinates(feature_type,
                                                                                            feature_coordinates,
                                                                                            protein,
                                                                                            settings)
                else:
                    projected_coordinates = None
                    if projected_coordinates:
                        vector = _PharmacophoreFeature.get_vector(projected_coordinates, feature_coordinates)
            else:
                projected = False

        return _PharmacophoreFeature(projected, feature_type, feature_coordinates, projected_coordinates, score, vector,
                                     settings)

    @staticmethod
    def from_crossminer(feature_str):
        """
        create feature from CrossMiner string

        :param str feature_str: extracted string from CrossMiner file
        :return: :class:`hotspots.hs_pharmacophore._PharmacophoreFeature`
        """
        cm_feature_dict = {"ring": "apolar",
                           "ring_planar_projected": "apolar",
                           "ring_non_planar": "apolar",
                           "hydrophobic": "apolar",
                           "donor_ch_projected": "donor",
                           "donor_projected": "donor",
                           "donor": "donor",
                           "acceptor_projected": "acceptor",
                           "acceptor": "acceptor",
                           "negative": "",
                           "positive": "",
                           }

        vector = None
        settings = PharmacophoreModel.Settings()

        feat = re.search(r"""PHARMACOPHORE_FEATURE (.+?)\n""", feature_str)
        if feat.group(1) == "excluded_volume":
            pass
        else:
            feature_type =  cm_feature_dict[feat.group(1)]

            spher = re.findall("""PHARMACOPHORE_SPHERE (.+?)\n""", feature_str)

            if len(spher) == 1:
                coords = spher[0].split(" ")
                feature_coordinates = Coordinates(float(coords[0]), float(coords[1]), float(coords[2]))
                projected = False
                projected_coordinates = None
                score = coords[3]

            elif len(spher) == 2:
                coords = spher[0].split(" ")
                proj =  spher[1].split(" ")
                feature_coordinates = Coordinates(float(coords[0]), float(coords[1]), float(coords[2]))
                projected = True
                projected_coordinates = Coordinates(float(proj[0]), float(proj[1]), float(proj[2]))
                score = coords[3]

            else:
                raise IOError("feature format not recognised")

            return _PharmacophoreFeature(projected, feature_type, feature_coordinates, projected_coordinates, score, vector,
                                         settings)

    @staticmethod
    def get_vector(projected_coordinates, feature_coordinates):
        """
        generates vector to hydrogen bonding partner on a protein.
        :return: Coordinates (named tuple)
        """
        return Coordinates(projected_coordinates.x - feature_coordinates.x,
                           projected_coordinates.y - feature_coordinates.y,
                           projected_coordinates.z - feature_coordinates.z)

    @staticmethod
    def get_projected_coordinates(feature_type, feature_coordinates, protein, settings):
        """
        for a given polar feature, the nearest h-bonding partner on the protein is located.
        :param protein: a :class:`ccdc.protein.Protein` instance
        :return: feature_coordinates for hydrogen-bonding partner
        """
        if feature_type == 'donor':
            atms = [a for a in protein.atoms if a.is_acceptor]
        else:
            atms = [a for a in protein.atoms if a.is_donor]

        near_atoms = {}
        for atm in atms:
            dist = Helper.get_distance(atm.coordinates, feature_coordinates)
            if dist < settings.max_hbond_dist:
                if dist in near_atoms.keys():
                    near_atoms[dist].append(atm)
                else:
                    near_atoms.update({dist:[atm]})
            else:
                continue
        if len(near_atoms.keys()) == 0:
            return None

        else:
            closest = sorted(near_atoms.keys())[0]
            select = near_atoms[closest][0]
            return select.coordinates

    @staticmethod
    def get_maxima(grid):
        """
        given a grid will return the max point
        :param grid:
        :return:
        """
        max_value = grid.extrema[1]
        indices = grid.indices_at_value(max_value)

        if len(indices) == 1:
            coords = grid.indices_to_point(indices[0][0], indices[0][1], indices[0][2])
            return max_value, Coordinates(coords[0], coords[1], coords[2])
        else:
            coords = grid.indices_to_point(round(sum(i[0] for i in indices)/len(indices)),
                                           round(sum(j[1] for j in indices)/len(indices)),
                                           round(sum(k[2] for k in indices)/len(indices))
                                           )
            return max_value, Coordinates(coords[0], coords[1], coords[2])

    @staticmethod
    def get_centroid(grid):
        """
        given a grid will return the centre of mass(mass ==HS score) and the score at that point
        :return: score, float, and coords, ccdc.molecule.Coordinate
        """

        weighted_x = 0
        weighted_y = 0
        weighted_z = 0
        total_mass = 0
        nx, ny, nz = grid.nsteps

        for i in range(nx):
            for j in range(ny):
                for k in range(nz):
                    grid_val = grid.value(i, j, k)
                    x, y, z = grid.indices_to_point(i, j, k)
                    weighted_x += grid_val * x
                    weighted_y += grid_val * y
                    weighted_z += grid_val * z
                    total_mass += grid_val

        coords = Coordinates(np.divide(weighted_x, total_mass),
                                      np.divide(weighted_y, total_mass),
                                      np.divide(weighted_z, total_mass)
                                      )
        score = grid.value_at_point(coords)

        return float(score), coords



"""
The :mod:`hotspots.hs_io` module was created to facilitate easy reading and
writing of Fragment Hotspot Map results.

There are multiple components to a :class:`hotspots.result.Result` including, the
protein, interaction grids and buriedness grid. It is therefore tedious to manually
read/write using the various class readers/writers. The Hotspots I/O organises this
for the user and can handle single :class:`hotspots.result.Result` or lists of
:class:`hotspots.result.Result`.


The main classes of the :mod:`hotspots.io` module are:

- :class:`hotspots.io.HotspotWriter`
- :class:`hotspots.io.HotspotReader`

"""
from __future__ import print_function

import shutil
import tempfile
import zipfile
from os import listdir
from os.path import splitext, join, basename, dirname, isdir

from ccdc import io
from ccdc.protein import Protein
from grid_extension import Grid
from hotspots.result import Results
from hotspots.hs_utilities import Helper
from hotspots.template_strings import pymol_imports, pymol_arrow, pymol_protein, pymol_grids, pymol_display_settings, \
    pymol_load_zip, pymol_labels, pymol_mesh


class HotspotWriter(Helper):
    """
    A class to handle the writing of a :class`hotspots.result.Result`. Additionally, creation of the
    PyMol visualisation scripts are handled here.

    :param str path: path to output directory
    :param str visualisation: "pymol" or "ngl" currently only PyMOL available
    :param str grid_extension: ".grd", ".ccp4" and ".acnt" supported
    :param bool zip_results: If True, the result directory will be compressed. (recommended)
    :param `hotspots.hs_io.HotspotWriter.Settings` settings: settings
    """

    class Settings(object):
        """
        A class to hold the :class:`hotspots.hs_io.HotspotWriter` settings
        """

        def __init__(self):
            self.grid_extension = ".grd"
            self.bg_color = "white"
            self.surface = False
            self.surface_trim_factor = 13
            self.organic_sticks = True
            self.isosurface_threshold = [10, 14, 17]
            self.grids = None
            self.transparency = 0.2
            self.grid_labels = True
            self.supported_grids = [".grd", ".ccp4", ".acnt"]
            self.pharmacophore = False
            self.pharmacophore_labels = True
            self.pharmacophore_format = [".py"]

    def __init__(self, path, visualisation="pymol", grid_extension=".grd", zip_results=True, settings=None):
        if settings is None:
            self.settings = self.Settings()
        else:
            self.settings = settings

        if grid_extension in self.settings.supported_grids:
            self.settings.grid_extension = grid_extension
        else:
            self.settings.grid_extension = ".grd"
            print("WARNING: Invalid grid file format provided. Default: '.grd' will be used")

        self.settings.visualisation = visualisation
        if visualisation == "ngl":
            self.settings.visualisation = visualisation
            if grid_extension != ".ccp4":
                self.settings.grid_extension = ".ccp4"
                print("WARNING: Grids must be .ccp4 format for visualisation with NGLViewer")

        self.path = self.get_out_dir(path)
        self.zipped = zip_results

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        if traceback:
            print(traceback)

    def write(self, hr):
        """
        writes the Fragment Hotspot Maps result to the output directory and create the pymol visualisation file

        :param `hotspots.result.Result` hr: a Fragment Hotspot Maps result or list of results

        >>> from hotspots.calculation import Runner
        >>> from hotspots.hs_io import HotspotWriter

        >>> r = Runner
        >>> result = r.from_pdb("1hcl")
        >>> out_dir = <path_to_out>
        >>> with HotspotWriter(out_dir) as w:
        >>>     w.write(result)


        """
        if isinstance(hr, list):
            self.settings.grids = hr[0].super_grids.keys()
            self.container = "hotspot_boundaries"
            self.number_of_hotspots = len(hr)

            self.out_dir = Helper.get_out_dir(join(self.path, self.container))

            self._write_protein(hr[0].protein)
            if hr[0].pharmacophore:
                self.settings.pharmacophore = True
            # hts = [h.hotspot_result for h in hr]
            self._write_pymol(hr, self.zipped)

            for i, hotspot in enumerate(hr):
                self.out_dir = Helper.get_out_dir(join(self.path, self.container, str(i)))
                self.settings.isosurface_threshold = [round(hotspot.threshold, 1)]

                bi = (Grid.super_grid(2, hotspot.best_island).max_value_of_neighbours()
                      > hotspot.threshold)

                self._write_grids(hotspot.super_grids, buriedness=None, mesh=bi)
                self._write_protein(hotspot.protein)

                if hotspot.pharmacophore:
                    self._write_pharmacophore(hotspot.pharmacophore)

                self._write_pymol(hotspot, False)

            self.out_dir = dirname(self.out_dir)
            if self.zipped:
                self.compress(join(dirname(self.out_dir), self.container))

        else:
            self.settings.grids = hr.super_grids.keys()
            self.container = "out"
            self.number_of_hotspots = 1

            self.out_dir = Helper.get_out_dir(join(self.path, self.container))
            self._write_grids(hr.super_grids, buriedness=hr.buriedness)
            self._write_protein(hr.protein)

            if hr.pharmacophore:
                self.settings.pharmacophore = True
                self._write_pharmacophore(hr.pharmacophore)
            self._write_pymol(hr, self.zipped)

            if self.zipped:
                self.compress(join(dirname(self.out_dir), self.container))

    def _write_grids(self, grid_dict, buriedness=None, mesh=None, out_dir=None):
        """
        writes grids to output directory
        :param grid_dict:
        :param buriedness:
        :return:
        """
        for p, g in grid_dict.items():
            fname = "{}{}".format(p, self.settings.grid_extension)
            if not out_dir:
                g.write(join(self.out_dir, fname))
            else:
                g.write(join(out_dir, fname))

        if buriedness:
            buriedness.write(join(self.out_dir, "buriedness{}".format(self.settings.grid_extension)))

        if mesh:
            mesh.write(join(self.out_dir, "mesh{}".format(self.settings.grid_extension)))

        if self.settings.grid_labels:
            labels = {"label_threshold_{}.mol2".format(threshold): self.get_label(grid_dict, threshold=threshold)
                      for threshold in self.settings.isosurface_threshold}

            for fname, label in labels.items():
                with io.MoleculeWriter(join(self.out_dir, fname)) as writer:
                    writer.write(label)

    def _write_protein(self, prot, out_dir=None):
        """
        writes protein to output directory
        :param prot:
        :return:
        """
        if not out_dir:
            out = join(self.out_dir, "protein.pdb")

        else:
            out = join(out_dir, "protein.pdb")

        with io.MoleculeWriter(out) as writer:
            writer.write(prot)

    def _write_pharmacophore(self, pharmacophore):
        """
        writes pharmacophore to output directory
        :param pharmacophore:
        :return:
        """
        out = [join(self.out_dir, "pharmacophore" + fmat) for fmat in self.settings.pharmacophore_format]
        for o in out:
            pharmacophore.write(o)

        label = self.get_label(pharmacophore)
        with io.MoleculeWriter(join(self.out_dir, "label_threshold_{}.mol2".format(pharmacophore.identifier))) \
                as writer:
            writer.write(label)

    def _write_pymol(self, hr, zipped=False):
        """
        Constructs PyMol python script to automatically display FH results
        :return:
        """
        pymol_out = pymol_imports()

        if self.settings.pharmacophore:
            pymol_out += pymol_arrow()

        if zipped:
            pymol_out += pymol_load_zip(basename(self.out_dir))

        pymol_out += pymol_protein(self.settings, self.zipped)

        if isinstance(hr, list):
            for i, h in enumerate(hr):
                self.settings.isosurface_threshold = [round(h.threshold, 1)]
                pymol_out += self._get_pymol_hotspot(h, i=i)
                pymol_out += pymol_mesh(i)

        else:
            pymol_out += self._get_pymol_hotspot(hr)

        pymol_out += pymol_display_settings(self.settings)

        if zipped:
            out = dirname(self.out_dir)
        else:
            out = self.out_dir

        with open('{}/pymol_results_file.py'.format(out), 'w') as pymol_file:
            pymol_file.write(pymol_out)

    def _get_pymol_hotspot(self, h, i=None):
        """
        core pymol sequence
        :param h: HotspotResult object
        :return:
        """
        pymol_out = ""
        if self.settings.grid_labels:

            for t in self.settings.isosurface_threshold:
                if i is not None:
                    fname = join(str(i), "label_threshold_{}.mol2".format(t))
                else:
                    fname = "label_threshold_{}.mol2".format(t)
                pymol_out += pymol_labels(fname=fname,
                                          objname="label_threshold_{}".format(t))

        pymol_out += pymol_grids(i, self.settings)

        if h.pharmacophore:
            if i is not None:
                f = join(str(i), "label_threshold_{}.mol2".format(h.pharmacophore.identifier))
            else:
                f = "label_threshold_{}.mol2".format(h.pharmacophore.identifier)

            pymol_out += h.pharmacophore._get_pymol_pharmacophore(lfile=f)

        return pymol_out

    def compress(self, archive_name, delete_directory=True):
        """
        compresses the output directory created for this :class:`hotspots.HotspotResults` instance, and
        removes the directory by default. The zipped file can be loaded directly into a new
        :class:`hotspots.HotspotResults` instance using the
        :func:`~hotspots.Hotspots.from_zip_dir` function

        :param str archive_name: file path
        :param bool delete_directory: remove the out directory once it has been zipped
        """
        self.archive_name = archive_name
        shutil.make_archive(self.archive_name, 'zip', self.out_dir)
        self.archive_loc = dirname("{}.zip".format(self.archive_name))
        if delete_directory:
            shutil.rmtree(self.out_dir)


class HotspotReader(object):
    """
    A class to organise the reading of a :class:`hotspots.result.Result`

    :param str path: path to the result directory (can be .zip directory)
    """

    def __init__(self, path):
        self._supported_interactions = ["apolar", "donor", "acceptor", "positive", "negative"]
        self._supported_grids = [".grd", ".ccp4", ".acnt", ".dat"]
        self._not_hs_dir = ["best_islands", "peaks", "ins"]
        self._path = path

        ext = splitext(self._path)[1]
        if ext == ".zip":
            self._base = self._path_from_zip()
        else:
            self._base = path

        self._files = listdir(self._base)
        self._extensions = set([splitext(f)[1] for f in self._files if f != "" or f != ".py"])

        pfiles = [f for f in self._files if splitext(f)[1] == ".pdb"]

        if len(pfiles) > 1:
            print("WARNING! {} has been used as default protein".format(join(self._base, "protein.pdb")))
            pfiles = [p for p in self._files if f == "protein.pdb"]

        self.protein = Protein.from_file(join(self._base, pfiles[0]))
        self.hs_dir = [d for d in self._files
                       if isdir(join(self._base, d)) and d not in self._not_hs_dir]

    def _path_from_zip(self):
        """
        writes files to temp dir, returns base dir
        :param self:
        :return:
        """
        base = tempfile.mkdtemp()

        with zipfile.ZipFile(self._path) as hs_zip:
            hs_zip.extractall(base)
        # d = splitext(basename(self._path))

        return base

    def _get_grids(self, sub_dir=None):
        """
        create a grid dictorionary
        :return:
        """
        if sub_dir:
            base = join(self._base, sub_dir)
            self._files = listdir(base)
            self._extensions = set([splitext(f)[1] for f in self._files if f != '' or f != '.py'])
        else:
            base = self._base

        if ".dat" in self._extensions:
            grid_dic = {splitext(fname)[0]: Grid.from_array(join(base, fname))
                        for fname in [f for f in self._files
                                      if splitext(f)[1] == ".grd"
                                      and splitext(f)[0] in self._supported_interactions]}
            try:
                buriedness = Grid.from_array(join(self.base, "buriedness.dat"))
            except RuntimeError:
                buriedness = None

        else:
            ext = list(set(self._extensions).intersection(self._supported_grids))
            if len(ext) == 1:
                grid_dic = {splitext(fname)[0]: Grid.from_file(join(base, fname))
                            for fname in [f for f in self._files
                                          if splitext(f)[1] == ext[0]
                                          and splitext(f)[0] in self._supported_interactions]}
                try:
                    buriedness = Grid.from_file("buriedness{}".format(ext[0]))
                except RuntimeError:
                    buriedness = None
            else:
                raise RuntimeError("Opps, something went wrong.")

        return grid_dic, buriedness

    def read(self, identifier=None):
        """
        creates a single or list of :class:`hotspots.result.Result` instance(s)

        :param str identifier: for directories containing multiple Fragment Hotspot Map results,
        identifier is the subdirectory for which a :class:`hotspots.result.Result` is requried

        :return: `hotspots.result.Result` a Fragment Hotspot Map result

        >>> from hotspots.hs_io import HotspotReader

        >>> path = "<path_to_results_directory>"
        >>> result = HotspotReader(path).read()


        """
        if len(self.hs_dir) == 0:
            self.grid_dic, self.buriedness = self._get_grids()
            return Results(protein=self.protein,
                           super_grids=self.grid_dic,
                           buriedness=self.buriedness)

        else:
            hrs = []
            if identifier:
                self.grid_dic, self.buriedness = self._get_grids(sub_dir=str(identifier))
                return Results(protein=self.protein,
                               super_grids=self.grid_dic,
                               buriedness=self.buriedness)
            else:
                for dir in self.hs_dir:
                    self.grid_dic, self.buriedness = self._get_grids(sub_dir=dir)
                    hrs.append(Results(protein=self.protein,
                                       super_grids=self.grid_dic,
                                       buriedness=self.buriedness))
            return hrs
